import datetime
import itertools
from contextlib import nullcontext
from unittest.mock import Mock

import pytest
from freezegun import freeze_time

from streamlink.stream.dash.manifest import MPD, DASHSegment, MPDParsers, MPDParsingError, Representation
from streamlink.utils.times import fromtimestamp
from tests.resources import xml


EPOCH_START = fromtimestamp(0)
UTC = datetime.timezone.utc


class TestSegment:
    @pytest.mark.parametrize(
        ("segmentdata", "expected"),
        [
            ({"uri": "https://foo/bar", "num": 123, "duration": 0.0, "init": True, "content": False}, "initialization"),
            ({"uri": "https://foo/bar", "num": 123, "duration": 0.0, "init": True, "content": True}, "123"),
            ({"uri": "https://foo/bar", "num": -1, "duration": 0.0, "init": True, "content": True}, "bar"),
            ({"uri": "https://foo/bar", "num": 123, "duration": 0.0}, "123"),
            ({"uri": "https://foo/bar", "num": -1, "duration": 0.0}, "bar"),
            ({"uri": "https://foo/bar/", "num": -1, "duration": 0.0}, "bar"),
            ({"uri": "https://foo/bar/baz.qux", "num": -1, "duration": 0.0}, "baz.qux"),
            ({"uri": "https://foo/bar/baz.qux?asdf", "num": -1, "duration": 0.0}, "baz.qux"),
        ],
    )
    def test_name(self, segmentdata: dict, expected: str):
        segment = DASHSegment(**segmentdata)
        assert segment.name == expected

    @pytest.mark.parametrize(
        ("available_at", "expected"),
        [
            (datetime.datetime(2000, 1, 2, 3, 4, 5, 123456, tzinfo=UTC), 1 * 86400 + 3 * 3600 + 4 * 60 + 5 + 0.123456),
            (datetime.datetime(2000, 1, 1, 0, 0, 0, 0, tzinfo=UTC), 0.0),
            (datetime.datetime(1999, 12, 31, 23, 59, 59, 999999, tzinfo=UTC), 0.0),
        ],
    )
    def test_available_in(self, available_at: datetime.datetime, expected: float):
        segment = DASHSegment(
            uri="foo",
            num=-1,
            duration=0.0,
            available_at=available_at,
        )
        with freeze_time("2000-01-01T00:00:00Z"):
            assert segment.available_in == pytest.approx(expected)

    def test_availability(self):
        segment = DASHSegment(
            uri="foo",
            num=-1,
            duration=0.0,
            available_at=datetime.datetime(2000, 1, 2, 3, 4, 5, 123456, tzinfo=UTC),
        )
        with freeze_time("2000-01-01T00:00:00Z"):
            assert segment.availability == "2000-01-02T03:04:05.123456Z / 2000-01-01T00:00:00.000000Z"


class TestMPDParsers:
    def test_bool_str(self):
        assert MPDParsers.bool_str("true")
        assert MPDParsers.bool_str("TRUE")
        assert MPDParsers.bool_str("True")

        assert not MPDParsers.bool_str("0")
        assert not MPDParsers.bool_str("False")
        assert not MPDParsers.bool_str("false")
        assert not MPDParsers.bool_str("FALSE")

    def test_type(self):
        assert MPDParsers.type("dynamic") == "dynamic"
        assert MPDParsers.type("static") == "static"
        with pytest.raises(MPDParsingError):
            # noinspection PyTypeChecker
            MPDParsers.type("other")

    def test_duration(self):
        assert MPDParsers.duration()("PT1S") == datetime.timedelta(seconds=1)
        assert MPDParsers.duration()("P1W") == datetime.timedelta(seconds=7 * 24 * 3600)

        assert MPDParsers.duration(EPOCH_START)("P3M") == datetime.timedelta(days=31 + 28 + 31)
        assert MPDParsers.duration(EPOCH_START + datetime.timedelta(31))("P3M") == datetime.timedelta(days=28 + 31 + 30)

        assert MPDParsers.duration(EPOCH_START)("P3Y") == datetime.timedelta(days=3 * 365 + 1)
        assert MPDParsers.duration(EPOCH_START + datetime.timedelta(3 * 365))("P3Y") == datetime.timedelta(days=3 * 365)

        with freeze_time(EPOCH_START):
            assert MPDParsers.duration()("P3M") == datetime.timedelta(days=31 + 28 + 31)
            assert MPDParsers.duration()("P3Y") == datetime.timedelta(days=3 * 365 + 1)

    def test_datetime(self):
        assert MPDParsers.datetime("2018-01-01T00:00:00Z") == datetime.datetime(2018, 1, 1, 0, 0, 0, tzinfo=UTC)

    def test_segment_template(self):
        assert MPDParsers.segment_template("$Time$-$Number$-$Other$")(Time=1, Number=2, Other=3) == "1-2-3"
        assert MPDParsers.segment_template("$Number%05d$")(Number=123) == "00123"
        assert MPDParsers.segment_template("$Time%0.02f$")(Time=100.234) == "100.23"

    def test_frame_rate(self):
        assert MPDParsers.frame_rate("1/25") == pytest.approx(1.0 / 25.0)
        assert MPDParsers.frame_rate("0.2") == pytest.approx(0.2)

    def test_timedelta(self):
        assert MPDParsers.timedelta(1)(100) == datetime.timedelta(0, 100.0)
        assert MPDParsers.timedelta(10)(100) == datetime.timedelta(0, 10.0)

    def test_range(self):
        assert MPDParsers.range("100-") == (100, None)
        assert MPDParsers.range("100-199") == (100, 100)
        with pytest.raises(MPDParsingError):
            MPDParsers.range("100")


class TestMPDParser:
    @pytest.mark.parametrize(
        ("min_buffer_time", "expected"),
        [
            pytest.param("PT1S", 3.0, id="minBufferTime lower than suggestedPresentationDelay"),
            pytest.param("PT5S", 5.0, id="minBufferTime greater than suggestedPresentationDelay"),
        ],
    )
    def test_suggested_presentation_delay(self, min_buffer_time: str, expected: float):
        with xml("dash/test_suggested_presentation_delay.mpd") as mpd_xml:
            mpd_xml.attrib["minBufferTime"] = min_buffer_time
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd")
        assert mpd.suggestedPresentationDelay.total_seconds() == expected

    def test_no_segment_list_or_template(self):
        with xml("dash/test_no_segment_list_or_template.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd")
            segments = [
                {
                    "ident": representation.ident,
                    "mimeType": representation.mimeType,
                    "segments": [
                        (segment.uri, segment.num, segment.duration, segment.available_at, segment.init, segment.content)
                        for segment in itertools.islice(representation.segments(), 100)
                    ],
                }
                for adaptationset in mpd.periods[0].adaptationSets
                for representation in adaptationset.representations
                if representation.id in ("1", "5", "6")
            ]
        availability = datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=UTC)

        assert segments == [
            {
                "ident": (None, None, "1"),
                "mimeType": "audio/mp4",
                "segments": [("http://cdn1.example.com/7657412348.mp4", -1, 3256.0, availability, True, True)],
            },
            {
                "ident": (None, None, "5"),
                "mimeType": "application/ttml+xml",
                "segments": [("http://cdn1.example.com/796735657.xml", -1, 3256.0, availability, True, True)],
            },
            {
                "ident": (None, None, "6"),
                "mimeType": "video/mp4",
                "segments": [("http://cdn1.example.com/8563456473.mp4", -1, 3256.0, availability, True, True)],
            },
        ]

    def test_segments_number_time(self):
        with xml("dash/test_1.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")

            segments = mpd.periods[0].adaptationSets[0].representations[0].segments()
            init_segment = next(segments)
            assert init_segment.uri == "http://test.se/tracks-v3/init-1526842800.g_m4v"

            # suggested delay is 11 seconds, each segment is 5 seconds long - so there should be 3
            assert [segment.uri for segment in itertools.islice(segments, 5)] == [
                "http://test.se/tracks-v3/dvr-1526842800-698.g_m4v?t=3403000",
                "http://test.se/tracks-v3/dvr-1526842800-699.g_m4v?t=3408000",
                "http://test.se/tracks-v3/dvr-1526842800-700.g_m4v?t=3413000",
            ]

    def test_segments_static_number(self):
        with xml("dash/test_2.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")

            segments = mpd.periods[0].adaptationSets[3].representations[0].segments()
            init_segment = next(segments)
            assert init_segment.uri == "http://test.se/video/250kbit/init.mp4"

            video_segments = [segment.uri for segment in itertools.islice(segments, 100000)]
            assert len(video_segments) == 444
            assert video_segments[:5] == [
                "http://test.se/video/250kbit/segment_1.m4s",
                "http://test.se/video/250kbit/segment_2.m4s",
                "http://test.se/video/250kbit/segment_3.m4s",
                "http://test.se/video/250kbit/segment_4.m4s",
                "http://test.se/video/250kbit/segment_5.m4s",
            ]

    def test_segments_dynamic_time(self):
        with xml("dash/test_3.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")

            segments = mpd.periods[0].adaptationSets[0].representations[0].segments()
            init_segment = next(segments)
            assert init_segment.uri == "http://test.se/video-2800000-0.mp4?z32="

            # default suggested delay is 3 seconds, each segment is 4 seconds long - so there should be 1 segment
            assert [segment.uri for segment in itertools.islice(segments, 3)] == [
                "http://test.se/video-time=1525450872000-2800000-0.m4s?z32=",
            ]

    # access manifest one hour after its availabilityStartTime
    @pytest.mark.parametrize(
        ("frozen_time", "timestamp"),
        [
            pytest.param(
                freeze_time("2000-01-01T01:00:00Z"),
                None,
                id="Without explicit timestamp",
            ),
            pytest.param(
                nullcontext(),
                datetime.datetime(2000, 1, 1, 1, 0, 0, 0, tzinfo=UTC),
                id="With explicit timestamp",
            ),
        ],
    )
    def test_segments_dynamic_number(self, frozen_time, timestamp):
        with xml("dash/test_segments_dynamic_number.mpd") as mpd_xml, frozen_time:
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd")
            segments_iterator = mpd.periods[0].adaptationSets[0].representations[0].segments(timestamp=timestamp)
            stream_urls = [(segment.uri, segment.available_at) for segment in itertools.islice(segments_iterator, 4)]

        assert stream_urls == [
            (
                "http://test/hd-5-init.mp4",
                datetime.datetime(2000, 1, 1, 0, 1, 30, tzinfo=UTC),
            ),
            (
                "http://test/hd-5_000000793.mp4",
                datetime.datetime(2000, 1, 1, 0, 59, 15, tzinfo=UTC),
            ),
            (
                "http://test/hd-5_000000794.mp4",
                datetime.datetime(2000, 1, 1, 0, 59, 20, tzinfo=UTC),
            ),
            (
                "http://test/hd-5_000000795.mp4",
                datetime.datetime(2000, 1, 1, 0, 59, 25, tzinfo=UTC),
            ),
        ]

    def test_static_no_publish_time(self):
        with xml("dash/test_static_no_publish_time.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd")

        segments = mpd.periods[0].adaptationSets[1].representations[0].segments()
        segment_urls = [(segment.uri, segment.available_at) for segment in itertools.islice(segments, 4)]
        # ignores period start time in static manifests
        expected_availability = datetime.datetime(2020, 1, 1, 0, 0, 0, tzinfo=UTC)

        assert segment_urls == [
            ("http://test/dash/150633-video_eng=194000.dash", expected_availability),
            ("http://test/dash/150633-video_eng=194000-0.dash", expected_availability),
            ("http://test/dash/150633-video_eng=194000-2000.dash", expected_availability),
            ("http://test/dash/150633-video_eng=194000-4000.dash", expected_availability),
        ]

    @pytest.mark.parametrize(
        ("manifest", "expected_time"),
        [
            pytest.param("dash/test_dynamic_no_publish_time_timeline_with_start.mpd", "155994772214430", id="with-start"),
            pytest.param("dash/test_dynamic_no_publish_time_timeline_without_start.mpd", "4320000", id="without-start"),
        ],
    )
    def test_dynamic_no_publish_time_with_timeline(self, manifest: str, expected_time: str):
        with xml(manifest) as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd")

        segments = [
            (segment.uri, segment.num, segment.duration, segment.available_at, segment.init, segment.content)
            for segment in itertools.islice(mpd.periods[0].adaptationSets[0].representations[0].segments(), 100)
        ]
        assert segments == [
            (
                f"http://test/video_1920x1080_avc1-{expected_time}.m4s",
                9,
                6.0,
                datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=UTC),
                False,
                True,
            ),
        ]

    def test_segment_list(self):
        with xml("dash/test_segment_list.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd")

        segments = mpd.periods[0].adaptationSets[0].representations[0].segments()
        segment_urls = [(segment.uri, segment.available_at) for segment in itertools.islice(segments, 4)]
        # ignores period start time in static manifests
        expected_availability = datetime.datetime(2020, 1, 1, 0, 0, 0, tzinfo=UTC)

        assert segment_urls == [
            ("http://test/chunk_ctvideo_ridp0va0br4332748_cinit_mpd.m4s", expected_availability),
            ("http://test/chunk_ctvideo_ridp0va0br4332748_cn1_mpd.m4s", expected_availability),
            ("http://test/chunk_ctvideo_ridp0va0br4332748_cn2_mpd.m4s", expected_availability),
            ("http://test/chunk_ctvideo_ridp0va0br4332748_cn3_mpd.m4s", expected_availability),
        ]

    def test_dynamic_segment_list_continued(self, caplog: pytest.LogCaptureFixture):
        caplog.set_level("WARNING", "streamlink.stream.dash")

        # init manifest
        with xml("dash/test_dynamic_segment_list_p1.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd")
        segments_iterator = mpd.periods[0].adaptationSets[0].representations[0].segments(init=True)

        assert [(segment.uri, segment.num) for segment in segments_iterator] == [
            ("http://test/init.m4s", -1),
            ("http://test/13.m4s", 13),
            ("http://test/14.m4s", 14),
            ("http://test/15.m4s", 15),
        ], "Queues the init segment and the correct number of segments from the live-edge"
        assert mpd.timelines["0", "0", "0"] == 16, "Remembers the next segment number"
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == []

        # regular continuation
        with xml("dash/test_dynamic_segment_list_p2.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd", timelines=mpd.timelines)
        segments_iterator = mpd.periods[0].adaptationSets[0].representations[0].segments(init=False)

        assert [(segment.uri, segment.num) for segment in segments_iterator] == [
            ("http://test/16.m4s", 16),
            ("http://test/17.m4s", 17),
            ("http://test/18.m4s", 18),
        ], "All segments from the remembered segment number were queued"
        assert mpd.timelines["0", "0", "0"] == 19, "Remembers the next segment number"
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == []

        # regular continuation with a different offset
        with xml("dash/test_dynamic_segment_list_p3.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd", timelines=mpd.timelines)
        segments_iterator = mpd.periods[0].adaptationSets[0].representations[0].segments(init=False)

        assert [(segment.uri, segment.num) for segment in segments_iterator] == [
            ("http://test/19.m4s", 19),
            ("http://test/20.m4s", 20),
            ("http://test/21.m4s", 21),
            ("http://test/22.m4s", 22),
            ("http://test/23.m4s", 23),
            ("http://test/24.m4s", 24),
        ], "All segments from the remembered segment number were queued"
        assert mpd.timelines["0", "0", "0"] == 25, "Remembers the next segment number"
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == []

        # skipped multiple segments
        with xml("dash/test_dynamic_segment_list_p4.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd", timelines=mpd.timelines)
        segments_iterator = mpd.periods[0].adaptationSets[0].representations[0].segments(init=False)

        assert [(segment.uri, segment.num) for segment in segments_iterator] == [
            ("http://test/30.m4s", 30),
            ("http://test/31.m4s", 31),
            ("http://test/32.m4s", 32),
            ("http://test/33.m4s", 33),
            ("http://test/34.m4s", 34),
            ("http://test/35.m4s", 35),
            ("http://test/36.m4s", 36),
            ("http://test/37.m4s", 37),
            ("http://test/38.m4s", 38),
            ("http://test/39.m4s", 39),
        ], "All segments from the remembered segment number were queued"
        assert mpd.timelines["0", "0", "0"] == 40, "Remembers the next segment number"
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.stream.dash.manifest",
                "warning",
                "Skipped segments 25-29 after manifest reload. This is unsupported and will result in incoherent output data.",
            ),
        ]

        caplog.records.clear()

        # skipped single segment
        with xml("dash/test_dynamic_segment_list_p5.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd", timelines=mpd.timelines)
        segments_iterator = mpd.periods[0].adaptationSets[0].representations[0].segments(init=False)

        assert [(segment.uri, segment.num) for segment in segments_iterator] == [
            ("http://test/41.m4s", 41),
            ("http://test/42.m4s", 42),
            ("http://test/43.m4s", 43),
            ("http://test/44.m4s", 44),
            ("http://test/45.m4s", 45),
            ("http://test/46.m4s", 46),
            ("http://test/47.m4s", 47),
            ("http://test/48.m4s", 48),
            ("http://test/49.m4s", 49),
            ("http://test/50.m4s", 50),
        ], "All segments from the remembered segment number were queued"
        assert mpd.timelines["0", "0", "0"] == 51, "Remembers the next segment number"
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.stream.dash.manifest",
                "warning",
                "Skipped segment 40 after manifest reload. This is unsupported and will result in incoherent output data.",
            ),
        ]

    def test_dynamic_segment_list_no_duration(self):
        with xml("dash/test_dynamic_segment_list_no_duration.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd")

        segments_iterator = mpd.periods[0].adaptationSets[0].representations[0].segments(init=True)
        assert [segment.uri for segment in segments_iterator] == [
            "http://test/init.m4s",
            "http://test/13.m4s",
            "http://test/14.m4s",
            "http://test/15.m4s",
        ]

    def test_dynamic_timeline_continued(self):
        with xml("dash/test_dynamic_timeline_continued_p1.mpd") as mpd_xml_p1:
            mpd_p1 = MPD(mpd_xml_p1, base_url="http://test/", url="http://test/manifest.mpd")
            iter_segment_p1 = mpd_p1.periods[0].adaptationSets[0].representations[0].segments()
            segments_p1 = [
                (segment.uri, segment.num, segment.duration, segment.available_at)
                for segment in itertools.islice(iter_segment_p1, 100)
            ]  # fmt: skip

        assert segments_p1 == [
            ("http://test/video/init.mp4", -1, 0.0, datetime.datetime(2018, 1, 1, 1, 0, 0, tzinfo=UTC)),
            ("http://test/video/1006000.mp4", 7, 1.0, datetime.datetime(2018, 1, 1, 12, 59, 56, tzinfo=UTC)),
            ("http://test/video/1007000.mp4", 8, 1.0, datetime.datetime(2018, 1, 1, 12, 59, 57, tzinfo=UTC)),
            ("http://test/video/1008000.mp4", 9, 1.0, datetime.datetime(2018, 1, 1, 12, 59, 58, tzinfo=UTC)),
            ("http://test/video/1009000.mp4", 10, 1.0, datetime.datetime(2018, 1, 1, 12, 59, 59, tzinfo=UTC)),
            ("http://test/video/1010000.mp4", 11, 1.0, datetime.datetime(2018, 1, 1, 13, 0, 0, tzinfo=UTC)),
        ]

        # continue with the next manifest
        with xml("dash/test_dynamic_timeline_continued_p2.mpd") as mpd_xml_p2:
            mpd_p2 = MPD(mpd_xml_p2, base_url=mpd_p1.base_url, url=mpd_p1.url, timelines=mpd_p1.timelines)
            iter_segment_p2 = mpd_p2.periods[0].adaptationSets[0].representations[0].segments(init=False)
            segments_p2 = [
                (segment.uri, segment.num, segment.duration, segment.available_at)
                for segment in itertools.islice(iter_segment_p2, 100)
            ]  # fmt: skip

        assert segments_p2 == [
            ("http://test/video/1011000.mp4", 7, 1.0, datetime.datetime(2018, 1, 1, 13, 0, 1, tzinfo=UTC)),
            ("http://test/video/1012000.mp4", 8, 1.0, datetime.datetime(2018, 1, 1, 13, 0, 2, tzinfo=UTC)),
            ("http://test/video/1013000.mp4", 9, 1.0, datetime.datetime(2018, 1, 1, 13, 0, 3, tzinfo=UTC)),
            ("http://test/video/1014000.mp4", 10, 1.0, datetime.datetime(2018, 1, 1, 13, 0, 4, tzinfo=UTC)),
            ("http://test/video/1015000.mp4", 11, 1.0, datetime.datetime(2018, 1, 1, 13, 0, 5, tzinfo=UTC)),
        ]

    def test_tsegment_t_is_none_1895(self):
        """
        Verify the fix for https://github.com/streamlink/streamlink/issues/1895
        """
        with xml("dash/test_8.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")

            segments = mpd.periods[0].adaptationSets[0].representations[0].segments()
            init_segment = next(segments)
            assert init_segment.uri == "http://test.se/video-2799000-0.mp4?z32=CENSORED_SESSION"

            video_segments = [x.uri for x in itertools.islice(segments, 3)]
            assert video_segments == [
                "http://test.se/video-time=0-2799000-0.m4s?z32=CENSORED_SESSION",
                "http://test.se/video-time=4000-2799000-0.m4s?z32=CENSORED_SESSION",
                "http://test.se/video-time=8000-2799000-0.m4s?z32=CENSORED_SESSION",
            ]

    def test_bitrate_rounded(self):
        def mock_rep(bandwidth):
            node = Mock(
                tag="Representation",
                attrib={
                    "id": "test",
                    "bandwidth": bandwidth,
                    "mimeType": "video/mp4",
                },
            )
            node.findall.return_value = []

            root = Mock()
            root.parent = None
            period = Mock()
            period.parent = root
            aset = Mock()
            aset.parent = period

            return Representation(node, root=root, parent=aset, period=period)

        assert mock_rep(1.2 * 1000.0).bandwidth_rounded == pytest.approx(1.2)
        assert mock_rep(45.6 * 1000.0).bandwidth_rounded == pytest.approx(46.0)
        assert mock_rep(134.0 * 1000.0).bandwidth_rounded == pytest.approx(130.0)
        assert mock_rep(1324.0 * 1000.0).bandwidth_rounded == pytest.approx(1300.0)

    def test_duplicated_resolutions(self):
        """
        Verify the fix for https://github.com/streamlink/streamlink/issues/3365
        """
        with xml("dash/test_10.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")

            representations_0 = mpd.periods[0].adaptationSets[0].representations[0]
            assert representations_0.height == 804
            assert representations_0.bandwidth == pytest.approx(10000.0)
            representations_1 = mpd.periods[0].adaptationSets[0].representations[1]
            assert representations_1.height == 804
            assert representations_1.bandwidth == pytest.approx(8000.0)

    def test_segments_static_periods_duration(self):
        """
        Verify the fix for https://github.com/streamlink/streamlink/issues/2873
        """
        with xml("dash/test_11_static.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")
            duration = mpd.periods[0].duration.total_seconds()
            assert duration == pytest.approx(204.32)

    def test_segments_byterange(self):
        with xml("dash/test_segments_byterange.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd")

        segment_urls = [
            [(seg.uri, seg.init, seg.byterange) for seg in adaptationset.representations[0].segments()]
            for adaptationset in mpd.periods[0].adaptationSets
        ]  # fmt: skip

        assert segment_urls == [
            [
                ("http://test/video-frag.mp4", True, (36, 711)),
                ("http://test/video-frag.mp4", False, (747, 875371)),
                ("http://test/video-frag.mp4", False, (876118, 590796)),
                ("http://test/video-frag.mp4", False, (1466914, 487041)),
                ("http://test/video-frag.mp4", False, (1953955, 40698)),
            ],
            [
                ("http://test/audio-frag.mp4", True, (32, 592)),
                ("http://test/audio-frag.mp4", False, (624, 123576)),
                ("http://test/audio-frag.mp4", False, (124200, 126104)),
                ("http://test/audio-frag.mp4", False, (250304, 124062)),
                ("http://test/audio-frag.mp4", False, (374366, 471)),
            ],
        ]

    def test_baseurl_urljoin_no_trailing_slash(self):
        with xml("dash/test_baseurl_urljoin.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="https://foo/bar", url="https://test/manifest.mpd")

        segment_urls = [
            [
                (period.id, adaptationset.id, segment.uri)
                for segment in itertools.islice(representation.segments(), 2)
            ]
            for period in mpd.periods
            for adaptationset in period.adaptationSets
            for representation in adaptationset.representations
        ]  # fmt: skip
        assert segment_urls == [
            [
                ("empty-baseurl", "absolute-segments", "https://foo/absolute/init_video_5000kbps.m4s"),
                ("empty-baseurl", "absolute-segments", "https://foo/absolute/media_video_5000kbps-1.m4s"),
            ],
            [
                ("empty-baseurl", "relative-segments", "https://foo/relative/init_video_5000kbps.m4s"),
                ("empty-baseurl", "relative-segments", "https://foo/relative/media_video_5000kbps-1.m4s"),
            ],
            [
                ("baseurl-with-scheme", "absolute-segments", "https://host/absolute/init_video_5000kbps.m4s"),
                ("baseurl-with-scheme", "absolute-segments", "https://host/absolute/media_video_5000kbps-1.m4s"),
            ],
            [
                ("baseurl-with-scheme", "relative-segments", "https://host/path/relative/init_video_5000kbps.m4s"),
                ("baseurl-with-scheme", "relative-segments", "https://host/path/relative/media_video_5000kbps-1.m4s"),
            ],
            [
                ("absolute-baseurl", "absolute-segments", "https://foo/absolute/init_video_5000kbps.m4s"),
                ("absolute-baseurl", "absolute-segments", "https://foo/absolute/media_video_5000kbps-1.m4s"),
            ],
            [
                ("absolute-baseurl", "relative-segments", "https://foo/path/relative/init_video_5000kbps.m4s"),
                ("absolute-baseurl", "relative-segments", "https://foo/path/relative/media_video_5000kbps-1.m4s"),
            ],
            [
                ("relative-baseurl", "absolute-segments", "https://foo/absolute/init_video_5000kbps.m4s"),
                ("relative-baseurl", "absolute-segments", "https://foo/absolute/media_video_5000kbps-1.m4s"),
            ],
            [
                ("relative-baseurl", "relative-segments", "https://foo/path/relative/init_video_5000kbps.m4s"),
                ("relative-baseurl", "relative-segments", "https://foo/path/relative/media_video_5000kbps-1.m4s"),
            ],
        ]

    def test_baseurl_urljoin_with_trailing_slash(self):
        with xml("dash/test_baseurl_urljoin.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="https://foo/bar/", url="https://test/manifest.mpd")

        segment_urls = [
            [
                (period.id, adaptationset.id, segment.uri)
                for segment in itertools.islice(representation.segments(), 2)
            ]
            for period in mpd.periods
            for adaptationset in period.adaptationSets
            for representation in adaptationset.representations
        ]  # fmt: skip
        assert segment_urls == [
            [
                ("empty-baseurl", "absolute-segments", "https://foo/absolute/init_video_5000kbps.m4s"),
                ("empty-baseurl", "absolute-segments", "https://foo/absolute/media_video_5000kbps-1.m4s"),
            ],
            [
                ("empty-baseurl", "relative-segments", "https://foo/bar/relative/init_video_5000kbps.m4s"),
                ("empty-baseurl", "relative-segments", "https://foo/bar/relative/media_video_5000kbps-1.m4s"),
            ],
            [
                ("baseurl-with-scheme", "absolute-segments", "https://host/absolute/init_video_5000kbps.m4s"),
                ("baseurl-with-scheme", "absolute-segments", "https://host/absolute/media_video_5000kbps-1.m4s"),
            ],
            [
                ("baseurl-with-scheme", "relative-segments", "https://host/path/relative/init_video_5000kbps.m4s"),
                ("baseurl-with-scheme", "relative-segments", "https://host/path/relative/media_video_5000kbps-1.m4s"),
            ],
            [
                ("absolute-baseurl", "absolute-segments", "https://foo/absolute/init_video_5000kbps.m4s"),
                ("absolute-baseurl", "absolute-segments", "https://foo/absolute/media_video_5000kbps-1.m4s"),
            ],
            [
                ("absolute-baseurl", "relative-segments", "https://foo/path/relative/init_video_5000kbps.m4s"),
                ("absolute-baseurl", "relative-segments", "https://foo/path/relative/media_video_5000kbps-1.m4s"),
            ],
            [
                ("relative-baseurl", "absolute-segments", "https://foo/absolute/init_video_5000kbps.m4s"),
                ("relative-baseurl", "absolute-segments", "https://foo/absolute/media_video_5000kbps-1.m4s"),
            ],
            [
                ("relative-baseurl", "relative-segments", "https://foo/bar/path/relative/init_video_5000kbps.m4s"),
                ("relative-baseurl", "relative-segments", "https://foo/bar/path/relative/media_video_5000kbps-1.m4s"),
            ],
        ]

    def test_baseurl_urljoin_empty(self):
        with xml("dash/test_baseurl_urljoin.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="", url="https://test/manifest.mpd")

        segment_urls = [
            [
                (period.id, adaptationset.id, segment.uri)
                for segment in itertools.islice(representation.segments(), 2)
            ]
            for period in mpd.periods
            for adaptationset in period.adaptationSets
            for representation in adaptationset.representations
        ]  # fmt: skip
        assert segment_urls == [
            [
                ("empty-baseurl", "absolute-segments", "/absolute/init_video_5000kbps.m4s"),
                ("empty-baseurl", "absolute-segments", "/absolute/media_video_5000kbps-1.m4s"),
            ],
            [
                ("empty-baseurl", "relative-segments", "relative/init_video_5000kbps.m4s"),
                ("empty-baseurl", "relative-segments", "relative/media_video_5000kbps-1.m4s"),
            ],
            [
                ("baseurl-with-scheme", "absolute-segments", "https://host/absolute/init_video_5000kbps.m4s"),
                ("baseurl-with-scheme", "absolute-segments", "https://host/absolute/media_video_5000kbps-1.m4s"),
            ],
            [
                ("baseurl-with-scheme", "relative-segments", "https://host/path/relative/init_video_5000kbps.m4s"),
                ("baseurl-with-scheme", "relative-segments", "https://host/path/relative/media_video_5000kbps-1.m4s"),
            ],
            [
                ("absolute-baseurl", "absolute-segments", "/absolute/init_video_5000kbps.m4s"),
                ("absolute-baseurl", "absolute-segments", "/absolute/media_video_5000kbps-1.m4s"),
            ],
            [
                ("absolute-baseurl", "relative-segments", "/path/relative/init_video_5000kbps.m4s"),
                ("absolute-baseurl", "relative-segments", "/path/relative/media_video_5000kbps-1.m4s"),
            ],
            [
                ("relative-baseurl", "absolute-segments", "/absolute/init_video_5000kbps.m4s"),
                ("relative-baseurl", "absolute-segments", "/absolute/media_video_5000kbps-1.m4s"),
            ],
            [
                ("relative-baseurl", "relative-segments", "path/relative/init_video_5000kbps.m4s"),
                ("relative-baseurl", "relative-segments", "path/relative/media_video_5000kbps-1.m4s"),
            ],
        ]

    def test_baseurl_nested(self):
        with xml("dash/test_baseurl_nested.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="https://foo/", url="https://test/manifest.mpd")

        segment_urls = [
            [(segment.uri, segment.available_at) for segment in itertools.islice(representation.segments(), 2)]
            for adaptationset in mpd.periods[0].adaptationSets
            for representation in adaptationset.representations
        ]
        # ignores period start time in static manifests
        expected_availability = datetime.datetime(2020, 1, 1, 0, 0, 0, tzinfo=UTC)

        assert segment_urls == [
            [
                ("https://hostname/period/init_video_5000kbps.m4s", expected_availability),
                ("https://hostname/period/media_video_5000kbps-1.m4s", expected_availability),
            ],
            [
                ("https://hostname/period/representation/init_video_9000kbps.m4s", expected_availability),
                ("https://hostname/period/representation/media_video_9000kbps-1.m4s", expected_availability),
            ],
            [
                ("https://hostname/period/adaptationset/init_audio_128kbps.m4s", expected_availability),
                ("https://hostname/period/adaptationset/media_audio_128kbps-1.m4s", expected_availability),
            ],
            [
                ("https://hostname/period/adaptationset/representation/init_audio_256kbps.m4s", expected_availability),
                ("https://hostname/period/adaptationset/representation/media_audio_256kbps-1.m4s", expected_availability),
            ],
            [
                ("https://other/init_audio_320kbps.m4s", expected_availability),
                ("https://other/media_audio_320kbps-1.m4s", expected_availability),
            ],
        ]

    def test_timeline_ids(self):
        with xml("dash/test_timeline_ids.mpd") as mpd_xml, freeze_time("2000-01-01T00:00:00Z"):
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd")
            segment_urls = [
                [segment.uri for segment in itertools.islice(representation.segments(), 3)]
                for adaptationset in mpd.periods[0].adaptationSets
                for representation in adaptationset.representations
            ]
        assert segment_urls == [
            [
                "http://test/audio1/init.mp4",
                "http://test/audio1/t0.m4s",
                "http://test/audio1/t1.m4s",
            ],
            [
                "http://test/audio2/init.mp4",
                "http://test/audio2/t0.m4s",
                "http://test/audio2/t1.m4s",
            ],
            [
                "http://test/video1/init.mp4",
                "http://test/video1/t0.m4s",
                "http://test/video1/t1.m4s",
            ],
            [
                "http://test/video2/init.mp4",
                "http://test/video2/t0.m4s",
                "http://test/video2/t1.m4s",
            ],
        ]
        assert list(mpd.timelines.keys()) == [
            ("period-0", "0", "audio1"),
            ("period-0", "0", "audio2"),
            ("period-0", None, "video1"),
            ("period-0", None, "video2"),
        ]

    def test_get_representation(self):
        with xml("dash/test_timeline_ids.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test/", url="http://test/manifest.mpd")

        assert mpd.get_representation((None, None, "unknown")) is None
        assert mpd.get_representation((None, None, "audio1")) is None
        assert mpd.get_representation((None, "0", "audio1")) is None
        assert mpd.get_representation(("period-0", None, "audio1")) is None

        assert getattr(mpd.get_representation(("period-0", "0", "audio1")), "mimeType", None) == "audio/mp4"
        assert getattr(mpd.get_representation(("period-0", "0", "audio2")), "mimeType", None) == "audio/mp4"
        assert getattr(mpd.get_representation(("period-0", None, "video1")), "mimeType", None) == "video/mp4"
        assert getattr(mpd.get_representation(("period-0", None, "video2")), "mimeType", None) == "video/mp4"



from streamlink.plugins.indihometv import IndiHomeTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlIndiHomeTV(PluginCanHandleUrl):
    __plugin__ = IndiHomeTV

    should_match = [
        "https://www.indihometv.com/livetv/seatoday",
        "https://www.indihometv.com/livetv/transtv",
        "https://www.indihometv.com/tvod/seatoday/1680109200/1680111000/18328552/voa-our-voice",
    ]



from streamlink.plugins.tvp import TVP
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTVP(PluginCanHandleUrl):
    __plugin__ = TVP

    should_match_groups = [
        # live
        (
            ("live", "https://vod.tvp.pl/live,1/tvp-info,399699/2025-05-20/serwis-info-dzien,2086908"),
            {"channel_id": "399699", "show_id": "2086908"},
        ),
        (
            ("live", "https://vod.tvp.pl/live,1/tvp-world,399731/2025-05-20/interview-from-vilnius---ep-146,2088070"),
            {"channel_id": "399731", "show_id": "2088070"},
        ),
        # VOD
        (
            ("vod", "https://vod.tvp.pl/filmy-fabularne,136/podroz-ksiecia,852411"),
            {"vod_id": "852411"},
        ),
        (
            ("vod", "https://vod.tvp.pl/filmy-fabularne,136/bambi-opowiesc-lesna,2053604"),
            {"vod_id": "2053604"},
        ),
        # VOD episode
        (
            ("vod", "https://vod.tvp.pl/seriale,18/dzikie-korytarze-odcinki,2092682/odcinek-1,S01E01,2092684"),
            {"vod_id": "2092684"},
        ),
        # tvp.info
        (
            (
                "tvp_info",
                "https://www.tvp.info/ogladaj-na-zywo",
            ),
            {},
        ),
        (
            (
                "tvp_info",
                "https://www.tvp.info/86792059/kot-przemytnik-utknal-na-wieziennym-ogrodzeniu",
            ),
            {},
        ),
        # sport.tvp
        (
            (
                "tvp_sport",
                "https://sport.tvp.pl/86453177/snooker-polska-liga-snookera-top16-polfinal-i-final-zapis",
            ),
            {"stream_id": "86453177"},
        ),
    ]

    should_not_match = [
        "https://tvp.pl/",
        "https://stream.tvp.pl/",
        "https://vod.tvp.pl/",
        "https://tvpstream.vod.tvp.pl/",
    ]



from typing import NamedTuple

import pytest

from streamlink_cli.console.terminal import cut_text, term_width, text_width


@pytest.mark.parametrize(
    "expected",
    [
        pytest.param(144, id="non-windows", marks=pytest.mark.posix_only()),
        pytest.param(143, id="windows", marks=pytest.mark.windows_only()),
    ],
)
def test_term_width(monkeypatch: pytest.MonkeyPatch, expected: int):
    class TerminalSize(NamedTuple):
        columns: int
        rows: int

    monkeypatch.setattr("streamlink_cli.console.terminal.get_terminal_size", lambda: TerminalSize(columns=144, rows=42))

    assert term_width() == expected


@pytest.mark.parametrize(
    ("chars", "expected"),
    [
        ("ABCDEFGHIJ", 10),
        ("A你好世界こんにちは안녕하세요B", 30),
        ("·「」『』【】-=！@#￥%……&×（）", 30),  # noqa: RUF001
    ],
)
def test_text_width(chars, expected):
    assert text_width(chars) == expected


@pytest.mark.parametrize(
    ("prefix", "max_len", "expected"),
    [
        ("你好世界こんにちは안녕하세요CD", 10, "녕하세요CD"),
        ("你好世界こんにちは안녕하세요CD", 9, "하세요CD"),
        ("你好世界こんにちは안녕하세요CD", 23, "こんにちは안녕하세요CD"),
    ],
)
def test_cut_text(prefix, max_len, expected):
    assert cut_text(prefix, max_len) == expected



from __future__ import annotations

import ssl
from ssl import SSLContext
from unittest.mock import Mock, PropertyMock, call

import pytest
import requests
import requests_mock as rm
from requests.adapters import HTTPAdapter
from urllib3.response import HTTPResponse

from streamlink import Streamlink
from streamlink.exceptions import PluginError, StreamlinkDeprecationWarning
from streamlink.session.http import HTTPSession, SSLContextAdapter, TLSNoDHAdapter, TLSSecLevel1Adapter
from streamlink.session.http_useragents import DEFAULT


class TestUrllib3Overrides:
    @pytest.fixture(scope="class")
    def httpsession(self) -> HTTPSession:
        return HTTPSession()

    @pytest.mark.parametrize(
        ("url", "expected", "assertion"),
        [
            (
                "https://foo/bar%3F?baz%21",
                "https://foo/bar%3F?baz%21",
                "Keeps encoded reserved characters",
            ),
            (
                "https://foo/%62%61%72?%62%61%7A",
                "https://foo/bar?baz",
                "Decodes encoded unreserved characters",
            ),
            (
                "https://foo/bär?bäz",
                "https://foo/b%C3%A4r?b%C3%A4z",
                "Encodes other characters",
            ),
            (
                "https://foo/b%c3%a4r?b%c3%a4z",
                "https://foo/b%c3%a4r?b%c3%a4z",
                "Keeps percent-encodings with lowercase characters",
            ),
            (
                "https://foo/b%C3%A4r?b%C3%A4z",
                "https://foo/b%C3%A4r?b%C3%A4z",
                "Keeps percent-encodings with uppercase characters",
            ),
            (
                "https://foo/%?%",
                "https://foo/%25?%25",
                "Empty percent-encodings without valid encodings",
            ),
            (
                "https://foo/%0?%0",
                "https://foo/%250?%250",
                "Incomplete percent-encodings without valid encodings",
            ),
            (
                "https://foo/%zz?%zz",
                "https://foo/%25zz?%25zz",
                "Invalid percent-encodings without valid encodings",
            ),
            (
                "https://foo/%3F%?%3F%",
                "https://foo/%253F%25?%253F%25",
                "Empty percent-encodings with valid encodings",
            ),
            (
                "https://foo/%3F%0?%3F%0",
                "https://foo/%253F%250?%253F%250",
                "Incomplete percent-encodings with valid encodings",
            ),
            (
                "https://foo/%3F%zz?%3F%zz",
                "https://foo/%253F%25zz?%253F%25zz",
                "Invalid percent-encodings with valid encodings",
            ),
        ],
    )
    def test_encode_invalid_chars(self, httpsession: HTTPSession, url: str, expected: str, assertion: str):
        req = requests.Request(method="GET", url=url)
        prep = httpsession.prepare_request(req)
        assert prep.url == expected, assertion


class TestHTTPSession:
    def test_session_init(self):
        session = HTTPSession()
        assert session.headers.get("User-Agent") == DEFAULT
        assert session.timeout == 20.0
        assert "file://" in session.adapters.keys()

    def test_read_timeout(self, monkeypatch: pytest.MonkeyPatch):
        mock_sleep = Mock()
        mock_request = Mock(side_effect=requests.Timeout)
        monkeypatch.setattr("streamlink.session.http.time.sleep", mock_sleep)
        monkeypatch.setattr("streamlink.session.http.Session.request", mock_request)

        session = HTTPSession()
        with pytest.raises(PluginError, match=r"^Unable to open URL: http://localhost/"):
            session.get("http://localhost/", timeout=123, retries=3, retry_backoff=2, retry_max_backoff=5)

        assert mock_request.call_args_list == [
            call("GET", "http://localhost/", headers={}, params={}, timeout=123, proxies={}, allow_redirects=True),
            call("GET", "http://localhost/", headers={}, params={}, timeout=123, proxies={}, allow_redirects=True),
            call("GET", "http://localhost/", headers={}, params={}, timeout=123, proxies={}, allow_redirects=True),
            call("GET", "http://localhost/", headers={}, params={}, timeout=123, proxies={}, allow_redirects=True),
        ]
        assert mock_sleep.call_args_list == [
            call(2),
            call(4),
            call(5),
        ]

    @pytest.mark.parametrize("encoding", ["UTF-32BE", "UTF-32LE", "UTF-16BE", "UTF-16LE", "UTF-8"])
    def test_determine_json_encoding(self, recwarn: pytest.WarningsRecorder, encoding: str):
        data = "Hello world, Γειά σου Κόσμε, こんにちは世界".encode(encoding)  # noqa: RUF001
        assert HTTPSession.determine_json_encoding(data) == encoding
        assert [(record.category, str(record.message)) for record in recwarn.list] == [
            (StreamlinkDeprecationWarning, "Deprecated HTTPSession.determine_json_encoding() call"),
        ]

    @pytest.mark.parametrize(
        ("encoding", "override"),
        [
            ("utf-32-be", None),
            ("utf-32-le", None),
            ("utf-16-be", None),
            ("utf-16-le", None),
            ("utf-8", None),
            # With byte order mark (BOM)
            ("utf-16", None),
            ("utf-32", None),
            ("utf-8-sig", None),
            # Override
            ("utf-8", "utf-8"),
            ("cp949", "cp949"),
        ],
    )
    def test_json(self, monkeypatch: pytest.MonkeyPatch, encoding: str, override: str | None):
        mock_content = PropertyMock(return_value='{"test": "Α and Ω"}'.encode(encoding))  # noqa: RUF001
        monkeypatch.setattr("requests.Response.content", mock_content)

        res = requests.Response()
        res.encoding = override

        assert HTTPSession.json(res) == {"test": "Α and Ω"}  # noqa: RUF001

    @pytest.mark.parametrize(
        ("content_type", "encoding", "content", "expected"),
        [
            pytest.param(
                "text/html",
                None,
                b"B\xe4r",
                "ISO-8859-1",
                id="default-iso-8859-1-charset-for-text",
            ),
            pytest.param(
                "application/json",
                None,
                b"B\xc3\xa4r",
                "utf-8",
                id="default-utf-8-charset-for-json",
            ),
            pytest.param(
                'text/html; charset="ISO-8859-1"',
                None,
                b"B\xe4r",
                "ISO-8859-1",
                id="declared-iso-8859-1-charset",
            ),
            pytest.param(
                'text/html; charset="utf-8"',
                None,
                b"B\xc3\xa4r",
                "utf-8",
                id="declared-utf-8-charset",
            ),
            pytest.param(
                "text/html",
                "utf-8",
                b"B\xc3\xa4r",
                "utf-8",
                id="override-missing-charset",
            ),
            pytest.param(
                'text/html; charset="ISO-8859-1"',
                "utf-8",
                b"B\xc3\xa4r",
                "utf-8",
                id="override-incorrect-charset",
            ),
            pytest.param(
                'text/html; charset="utf-8"',
                "utf-8",
                b"B\xc3\xa4r",
                "utf-8",
                id="override-same-charset",
            ),
        ],
    )
    @pytest.mark.parametrize("method", ["get", "post", "head", "put", "patch", "delete"])
    def test_encoding_override(
        self,
        requests_mock: rm.Mocker,
        method: str,
        content_type: str,
        encoding: str,
        content: bytes,
        expected: str,
    ):
        requests_mock.register_uri(rm.ANY, "http://mocked", headers={"Content-Type": content_type}, content=content)
        httpsession = HTTPSession()
        res = getattr(httpsession, method)("http://mocked", encoding=encoding)
        assert res.encoding == expected
        assert res.text == "Bär"


class TestHTTPAdapters:
    @staticmethod
    def _has_dh_ciphers(ssl_context: SSLContext):
        return any(cipher["kea"] == "kx-dhe" for cipher in ssl_context.get_ciphers())

    @staticmethod
    def _has_weak_digest_ciphers(ssl_context: SSLContext):
        return any(cipher["digest"] == "sha1" for cipher in ssl_context.get_ciphers())

    def test_sslcontextadapter(self):
        adapter = SSLContextAdapter()
        ssl_context = adapter.poolmanager.connection_pool_kw.get("ssl_context")
        assert isinstance(ssl_context, SSLContext)
        assert self._has_dh_ciphers(ssl_context)
        assert not self._has_weak_digest_ciphers(ssl_context)

    def test_tlsnodhadapter(self):
        adapter = TLSNoDHAdapter()
        ssl_context = adapter.poolmanager.connection_pool_kw.get("ssl_context")
        assert isinstance(ssl_context, SSLContext)
        assert not self._has_dh_ciphers(ssl_context)
        assert not self._has_weak_digest_ciphers(ssl_context)

    def test_tlsseclevel1adapter(self):
        adapter = TLSSecLevel1Adapter()
        ssl_context = adapter.poolmanager.connection_pool_kw.get("ssl_context")
        assert isinstance(ssl_context, SSLContext)
        assert self._has_dh_ciphers(ssl_context)
        assert self._has_weak_digest_ciphers(ssl_context)

    @pytest.mark.parametrize("proxy", ["http", "socks4", "socks5"])
    def test_proxymanager_ssl_context(self, proxy: str):
        adapter = SSLContextAdapter()
        proxymanager = adapter.proxy_manager_for(f"{proxy}://")
        ssl_context_poolmanager = adapter.poolmanager.connection_pool_kw.get("ssl_context")
        ssl_context_proxymanager = proxymanager.connection_pool_kw.get("ssl_context")
        assert ssl_context_poolmanager is ssl_context_proxymanager


class TestHTTPSessionVerifyAndCustomSSLContext:
    @pytest.fixture()
    def adapter(self, session: Streamlink):
        # The http-disable-dh session option mounts the TLSNoDHAdapter with a custom SSLContext
        session.set_option("http-disable-dh", True)

        adapter = session.http.adapters.get("https://")
        assert isinstance(adapter, TLSNoDHAdapter)

        return adapter

    @pytest.fixture(autouse=True)
    def _fake_request(self, monkeypatch: pytest.MonkeyPatch, session: Streamlink, adapter: HTTPAdapter):
        class FakeHTTPResponse(HTTPResponse):
            def stream(self, *_, **__):
                yield b"mocked"

        # Can't use requests_mock here, as it monkeypatches the adapter's send() method, which we want to test
        req = requests.PreparedRequest()
        resp = FakeHTTPResponse(status=200)
        # noinspection PyTypeChecker
        response = adapter.build_response(req, resp)
        monkeypatch.setattr("requests.adapters.HTTPAdapter.send", Mock(return_value=response))

        assert session.http.get("https://mocked/").text == "mocked"

    @pytest.mark.parametrize(
        ("session", "check_hostname", "verify_mode"),
        [
            pytest.param({"http-ssl-verify": True}, True, ssl.CERT_REQUIRED, id="verify"),
            pytest.param({"http-ssl-verify": False}, False, ssl.CERT_NONE, id="no-verify"),
        ],
        indirect=["session"],
    )
    def test_ssl_context_attributes(
        self,
        session: Streamlink,
        adapter: HTTPAdapter,
        check_hostname: bool,
        verify_mode: ssl.VerifyMode,
    ):
        assert session.http.verify is session.get_option("http-ssl-verify")

        ssl_context = adapter.poolmanager.connection_pool_kw.get("ssl_context")
        assert isinstance(ssl_context, SSLContext)
        assert ssl_context.check_hostname is check_hostname
        assert ssl_context.verify_mode is verify_mode



from streamlink.plugins.vimeo import Vimeo
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlVimeo(PluginCanHandleUrl):
    __plugin__ = Vimeo

    should_match_groups = [
        (("default", "https://vimeo.com/783455878"), {}),
        (("default", "https://vimeo.com/channels/music/176894130"), {}),
        (("default", "https://vimeo.com/ondemand/worldoftomorrow3/467204924"), {}),
        (("default", "https://vimeo.com/ondemand/100footsurfingdays"), {}),
        (("player", "https://player.vimeo.com/video/176894130"), {}),
        (("event", "https://vimeo.com/event/4154130"), {"event_id": "4154130"}),
        (("event", "https://vimeo.com/event/4154130/embed"), {"event_id": "4154130"}),
    ]

    should_not_match = [
        "https://www.vimeo.com/",
    ]



from __future__ import annotations

import sys
from pathlib import PurePath, PurePosixPath, PureWindowsPath
from time import time
from unittest.mock import Mock

import freezegun
import pytest

from streamlink_cli.console.progress import Progress, ProgressFormatter
from tests.testutils.handshake import Handshake


class TestProgressFormatter:
    @pytest.fixture(scope="class")
    def params(self):
        return dict(
            written="WRITTEN",
            elapsed="ELAPSED",
            speed="SPEED",
            path=lambda *_: "PATH",
        )

    @pytest.fixture(autouse=True)
    def term_width(self, request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
        width = getattr(request, "param", 99)
        monkeypatch.setattr("streamlink_cli.console.progress.term_width", lambda: width)

    @pytest.mark.parametrize(
        ("term_width", "expected"),
        [
            (99, "[download] Written WRITTEN to PATH (ELAPSED @ SPEED)"),
            (63, "[download] Written WRITTEN to PATH (ELAPSED @ SPEED)"),
            (62, "[download] Written WRITTEN (ELAPSED @ SPEED)"),
            (44, "[download] Written WRITTEN (ELAPSED @ SPEED)"),
            (43, "[download] WRITTEN (ELAPSED @ SPEED)"),
            (36, "[download] WRITTEN (ELAPSED @ SPEED)"),
            (35, "[download] WRITTEN (ELAPSED)"),
            (28, "[download] WRITTEN (ELAPSED)"),
            (27, "[download] WRITTEN"),
            (1, "[download] WRITTEN"),
        ],
        indirect=["term_width"],
    )
    def test_format(self, params, term_width, expected):
        assert ProgressFormatter.format(ProgressFormatter.FORMATS, params) == expected

    @pytest.mark.parametrize(
        ("term_width", "expected"),
        [
            (99, "[download] Written WRITTEN to PATH (ELAPSED)"),
            (55, "[download] Written WRITTEN to PATH (ELAPSED)"),
            (54, "[download] Written WRITTEN (ELAPSED)"),
            (36, "[download] Written WRITTEN (ELAPSED)"),
            (35, "[download] WRITTEN (ELAPSED)"),
            (28, "[download] WRITTEN (ELAPSED)"),
            (27, "[download] WRITTEN"),
            (1, "[download] WRITTEN"),
        ],
        indirect=["term_width"],
    )
    def test_format_nospeed(self, params, term_width, expected):
        assert ProgressFormatter.format(ProgressFormatter.FORMATS_NOSPEED, params) == expected

    def test_format_missing(self, params):
        assert ProgressFormatter.format(ProgressFormatter.FORMATS, {"written": "0"}) == "[download] 0"

    def test_format_error(self, params):
        params = dict(**params)
        params["path"] = Mock(side_effect=ValueError("fail"))
        assert ProgressFormatter.format(ProgressFormatter.FORMATS, params) == "[download] Written WRITTEN (ELAPSED @ SPEED)"

    @pytest.mark.parametrize(
        ("size", "expected"),
        [
            (0, "0 bytes"),
            (2**10 - 1, "1023 bytes"),
            (2**10, "1.00 KiB"),
            (2**20 - 1, "1023.99 KiB"),
            (2**20, "1.00 MiB"),
            (2**30 - 1, "1023.99 MiB"),
            (2**30, "1.00 GiB"),
            (2**40 - 1, "1023.99 GiB"),
            (2**40, "1.00 TiB"),
            (2**50 - 1, "1023.99 TiB"),
            (2**50, "1024.00 TiB"),
        ],
    )
    def test_format_filesize(self, size, expected):
        assert ProgressFormatter.format_filesize(size) == expected
        assert ProgressFormatter.format_filesize(float(size)) == expected
        assert ProgressFormatter.format_filesize(size, "/s") == f"{expected}/s"

    @pytest.mark.parametrize(
        ("elapsed", "expected"),
        [
            (-1, "0s"),
            (0, "0s"),
            (9, "9s"),
            (10, "10s"),
            (59, "59s"),
            (60, "1m00s"),
            (69, "1m09s"),
            (70, "1m10s"),
            (119, "1m59s"),
            (120, "2m00s"),
            (3599, "59m59s"),
            (3600, "1h00m00s"),
            (3659, "1h00m59s"),
            (3660, "1h01m00s"),
            (3661, "1h01m01s"),
            (86399, "23h59m59s"),
            (86400, "24h00m00s"),
            (172800, "48h00m00s"),
        ],
    )
    def test_format_time(self, elapsed, expected):
        assert ProgressFormatter.format_time(elapsed) == expected


_PATH_POSIX = PurePosixPath("/foobar/baz/some file name")
_PATH_WIN_ABS = PureWindowsPath("C:\\foobar\\baz\\some file name")
_PATH_WIN_REL = PureWindowsPath("foobar\\baz\\some file name")
_PATH_WIN_UNC = PureWindowsPath("\\\\?\\foobar\\baz\\some file name")


class _TestFormatPath:
    # noinspection PyMethodMayBeStatic
    def test_format_path(self, monkeypatch: pytest.MonkeyPatch, path: PurePath, max_width: int, expected: str):
        monkeypatch.setattr("os.path.sep", "\\" if type(path) is PureWindowsPath else "/")
        assert ProgressFormatter.format_path(path, max_width) == expected


@pytest.mark.parametrize(
    ("path", "max_width", "expected"),
    [
        pytest.param(_PATH_POSIX, 26, "/foobar/baz/some file name", id="full path"),
        pytest.param(_PATH_POSIX, 25, "…oobar/baz/some file name", id="truncated by 1"),
        pytest.param(_PATH_POSIX, 24, "…obar/baz/some file name", id="truncated by 2"),
        pytest.param(_PATH_POSIX, 23, "…bar/baz/some file name", id="truncated by 3"),
        pytest.param(_PATH_POSIX, 22, "…ar/baz/some file name", id="truncated by 4"),
        pytest.param(_PATH_POSIX, 21, "…r/baz/some file name", id="truncated by 5"),
        pytest.param(_PATH_POSIX, 20, "…/baz/some file name", id="truncated by 6"),
        pytest.param(_PATH_POSIX, 19, "…baz/some file name", id="truncated by 7 (cuts off separator)"),
        pytest.param(_PATH_POSIX, 16, "…/some file name", id="truncated (all parts except name)"),
        pytest.param(_PATH_POSIX, 15, "…some file name", id="truncated (name without separator)"),
        pytest.param(_PATH_POSIX, 14, "…ome file name", id="truncated name"),
    ],
)
class TestFormatPathPOSIX(_TestFormatPath):
    pass


@pytest.mark.parametrize(
    ("path", "max_width", "expected"),
    [
        pytest.param(_PATH_WIN_ABS, 28, "C:\\foobar\\baz\\some file name", id="full path"),
        pytest.param(_PATH_WIN_ABS, 27, "C:…oobar\\baz\\some file name", id="truncated by 1"),
        pytest.param(_PATH_WIN_ABS, 26, "C:…obar\\baz\\some file name", id="truncated by 2"),
        pytest.param(_PATH_WIN_ABS, 25, "C:…bar\\baz\\some file name", id="truncated by 3"),
        pytest.param(_PATH_WIN_ABS, 24, "C:…ar\\baz\\some file name", id="truncated by 4"),
        pytest.param(_PATH_WIN_ABS, 23, "C:…r\\baz\\some file name", id="truncated by 5"),
        pytest.param(_PATH_WIN_ABS, 22, "C:…\\baz\\some file name", id="truncated by 6"),
        pytest.param(_PATH_WIN_ABS, 21, "C:…baz\\some file name", id="truncated by 7 (cuts off separator)"),
        pytest.param(_PATH_WIN_ABS, 18, "C:…\\some file name", id="truncated (all parts except name)"),
        pytest.param(_PATH_WIN_ABS, 17, "C:…some file name", id="truncated (name without separator)"),
        pytest.param(_PATH_WIN_ABS, 16, "C:…ome file name", id="truncated name"),
    ],
)
class TestFormatPathWindowsAbsolute(_TestFormatPath):
    pass


@pytest.mark.parametrize(
    ("path", "max_width", "expected"),
    [
        pytest.param(_PATH_WIN_REL, 25, "foobar\\baz\\some file name", id="full path"),
        pytest.param(_PATH_WIN_REL, 24, "…obar\\baz\\some file name", id="truncated by 1"),
        pytest.param(_PATH_WIN_REL, 23, "…bar\\baz\\some file name", id="truncated by 2"),
        pytest.param(_PATH_WIN_REL, 22, "…ar\\baz\\some file name", id="truncated by 3"),
        pytest.param(_PATH_WIN_REL, 21, "…r\\baz\\some file name", id="truncated by 4"),
        pytest.param(_PATH_WIN_REL, 20, "…\\baz\\some file name", id="truncated by 5"),
        pytest.param(_PATH_WIN_REL, 19, "…baz\\some file name", id="truncated by 6 (cuts off separator)"),
        pytest.param(_PATH_WIN_REL, 16, "…\\some file name", id="truncated (all parts except name)"),
        pytest.param(_PATH_WIN_REL, 15, "…some file name", id="truncated (name without separator)"),
        pytest.param(_PATH_WIN_REL, 14, "…ome file name", id="truncated name"),
    ],
)
class TestFormatPathWindowsRelative(_TestFormatPath):
    pass


@pytest.mark.parametrize(
    ("path", "max_width", "expected"),
    [
        # <py312: server/host name is not part of the path's drive, so it'll get truncated
        pytest.param(_PATH_WIN_UNC, 29, "\\\\?\\foobar\\baz\\some file name", id="full path"),
        pytest.param(_PATH_WIN_UNC, 28, "\\\\?\\…obar\\baz\\some file name", id="truncated by 1"),
        pytest.param(_PATH_WIN_UNC, 20, "\\\\?\\…\\some file name", id="truncated (all parts except name)"),
        pytest.param(_PATH_WIN_UNC, 19, "\\\\?\\…some file name", id="truncated (name without separator)"),
        pytest.param(_PATH_WIN_UNC, 18, "\\\\?\\…ome file name", id="truncated name"),
    ]
    if sys.version_info < (3, 12)
    else [
        # >=py312: server/host name is part of the path's drive, so it won't get truncated
        pytest.param(_PATH_WIN_UNC, 29, "\\\\?\\foobar\\baz\\some file name", id="full path"),
        pytest.param(_PATH_WIN_UNC, 28, "\\\\?\\foobar…az\\some file name", id="truncated by 1"),
        pytest.param(_PATH_WIN_UNC, 26, "\\\\?\\foobar…\\some file name", id="truncated (all parts except name)"),
        pytest.param(_PATH_WIN_UNC, 25, "\\\\?\\foobar…some file name", id="truncated (name without separator)"),
        pytest.param(_PATH_WIN_UNC, 24, "\\\\?\\foobar…ome file name", id="truncated name"),
    ],
)
class TestFormatPathWindowsUniversalNamingConvention(_TestFormatPath):
    pass


class TestProgress:
    @pytest.fixture(autouse=True)
    def _setup(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("os.path.sep", "/")

    @pytest.fixture(autouse=True)
    def mock_width(self, monkeypatch: pytest.MonkeyPatch):
        mock = Mock(return_value=70)
        monkeypatch.setattr("streamlink_cli.console.progress.term_width", mock)
        return mock

    @pytest.fixture()
    def frozen_time(self):
        with freezegun.freeze_time("2000-01-01T00:00:00Z") as frozen_time:
            yield frozen_time

    def test_download_speed(self, mock_width: Mock, frozen_time):
        kib = b"\x00" * 1024
        messages: list[str] = []
        console = Mock(msg_status=Mock(side_effect=messages.append))
        progress = Progress(
            console=console,
            path=PurePosixPath("../../the/path/where/we/write/to"),
            interval=1,
            history=3,
            threshold=2,
        )

        progress.started = time()
        assert messages == []

        progress.update()
        assert messages.pop() == "[download] Written 0 bytes to ../../the/path/where/we/write/to (0s)"

        progress.update()
        assert messages.pop() == "[download] Written 0 bytes to ../../the/path/where/we/write/to (0s)"

        frozen_time.tick()
        progress.write(kib * 1)
        progress.update()
        assert messages.pop() == "[download] Written 1.00 KiB to …th/where/we/write/to (1s @ 1.00 KiB/s)"

        frozen_time.tick()
        mock_width.return_value = 65
        progress.write(kib * 3)
        progress.update()
        assert messages.pop() == "[download] Written 4.00 KiB to …ere/we/write/to (2s @ 2.00 KiB/s)"

        frozen_time.tick()
        mock_width.return_value = 60
        progress.write(kib * 5)
        progress.update()
        assert messages.pop() == "[download] Written 9.00 KiB (3s @ 4.50 KiB/s)"

        frozen_time.tick()
        progress.write(kib * 7)
        progress.update()
        assert messages.pop() == "[download] Written 16.00 KiB (4s @ 7.50 KiB/s)"

        frozen_time.tick()
        progress.write(kib * 5)
        progress.update()
        assert messages.pop() == "[download] Written 21.00 KiB (5s @ 8.50 KiB/s)"

        frozen_time.tick()
        progress.update()
        assert messages.pop() == "[download] Written 21.00 KiB (6s @ 6.00 KiB/s)"

    @pytest.mark.parametrize(
        "status",
        [
            pytest.param(True, id="status"),
            pytest.param(False, id="no-status"),
        ],
    )
    def test_update(self, status: bool):
        handshake = Handshake()

        class _Progress(Progress):
            def update(self):
                with handshake():
                    return super().update()

        messages_status: list[str] = []
        messages_default: list[str] = []
        console = Mock(
            msg_status=Mock(side_effect=messages_status.append),
            msg=Mock(side_effect=messages_default.append),
        )
        thread = _Progress(console=console, path=PurePath(), status=status)
        # override the thread's polling time after initializing the deque of the rolling average download speed:
        # the interval constructor keyword is used to set the deque size
        thread.interval = 0
        thread.start()

        target = messages_status if status else messages_default
        other = messages_default if status else messages_status

        # first tick
        assert handshake.wait_ready(1)
        thread.write(b"123")
        assert handshake.step(1)
        assert target.pop().startswith("[download] Written 3 bytes")
        assert other == []

        # second tick
        assert handshake.wait_ready(1)
        thread.write(b"465")
        assert handshake.step(1)
        assert target.pop().startswith("[download] Written 6 bytes")
        assert other == []

        # close progress thread
        assert handshake.wait_ready(1)
        thread.close()
        assert handshake.step(1)
        assert target.pop().startswith("[download] Written 6 bytes")
        assert other == []

        # write data right after closing the thread, but before it has halted
        thread.write(b"789")
        handshake.go()
        thread.join(1)
        assert not thread.is_alive()
        assert target.pop().startswith("[download] Written 9 bytes")
        assert other == []



from __future__ import annotations

from contextlib import nullcontext
from pathlib import Path, PurePosixPath, PureWindowsPath
from signal import SIGTERM
from typing import Any

import pytest
import requests_mock as rm
import trio
from requests import Timeout

from streamlink.compat import is_win32
from streamlink.exceptions import PluginError
from streamlink.session import Streamlink
from streamlink.webbrowser.chromium import ChromiumWebbrowser
from streamlink.webbrowser.exceptions import WebbrowserError


class TestInit:
    @pytest.mark.parametrize(
        ("executable", "resolve_executable", "raises"),
        [
            pytest.param(
                None,
                None,
                pytest.raises(
                    WebbrowserError,
                    match=r"^Could not find Chromium-based web browser executable: Please set the path ",
                ),
                id="Failure with unset path",
            ),
            pytest.param(
                "custom",
                None,
                pytest.raises(
                    WebbrowserError,
                    match=r"^Invalid web browser executable: custom$",
                ),
                id="Failure with custom path",
            ),
            pytest.param(
                None,
                "default",
                nullcontext(),
                id="Success with default path",
            ),
            pytest.param(
                "custom",
                "custom",
                nullcontext(),
                id="Success with custom path",
            ),
        ],
        indirect=["resolve_executable"],
    )
    def test_resolve_executable(self, resolve_executable, executable: str | None, raises: nullcontext):
        with raises:
            ChromiumWebbrowser(executable=executable)


class TestFallbacks:
    def test_win32(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("streamlink.webbrowser.chromium.is_win32", True)
        monkeypatch.setattr("streamlink.webbrowser.chromium.is_darwin", False)
        monkeypatch.setattr("streamlink.webbrowser.chromium.Path", PureWindowsPath)
        monkeypatch.setattr(
            "os.getenv",
            {
                "PROGRAMFILES": "C:\\Program Files",
                "PROGRAMFILES(X86)": "C:\\Program Files (x86)",
                "LOCALAPPDATA": "C:\\Users\\user\\AppData\\Local",
            }.get,
        )
        assert ChromiumWebbrowser.fallback_paths() == [
            "C:\\Program Files\\Microsoft\\Edge\\Application\\msedge.exe",
            "C:\\Program Files (x86)\\Microsoft\\Edge\\Application\\msedge.exe",
            "C:\\Program Files\\Microsoft\\Edge Beta\\Application\\msedge.exe",
            "C:\\Program Files (x86)\\Microsoft\\Edge Beta\\Application\\msedge.exe",
            "C:\\Program Files\\Microsoft\\Edge Dev\\Application\\msedge.exe",
            "C:\\Program Files (x86)\\Microsoft\\Edge Dev\\Application\\msedge.exe",
            "C:\\Program Files\\Google\\Chrome\\Application\\chrome.exe",
            "C:\\Program Files (x86)\\Google\\Chrome\\Application\\chrome.exe",
            "C:\\Users\\user\\AppData\\Local\\Google\\Chrome\\Application\\chrome.exe",
            "C:\\Program Files\\Google\\Chrome Beta\\Application\\chrome.exe",
            "C:\\Program Files (x86)\\Google\\Chrome Beta\\Application\\chrome.exe",
            "C:\\Users\\user\\AppData\\Local\\Google\\Chrome Beta\\Application\\chrome.exe",
            "C:\\Program Files\\Google\\Chrome Canary\\Application\\chrome.exe",
            "C:\\Program Files (x86)\\Google\\Chrome Canary\\Application\\chrome.exe",
            "C:\\Users\\user\\AppData\\Local\\Google\\Chrome Canary\\Application\\chrome.exe",
        ]

    def test_darwin(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("streamlink.webbrowser.chromium.is_win32", False)
        monkeypatch.setattr("streamlink.webbrowser.chromium.is_darwin", True)
        monkeypatch.setattr("streamlink.webbrowser.chromium.Path", PurePosixPath)
        PurePosixPath.home = lambda: PurePosixPath("/Users/user")  # type: ignore[attr-defined]
        assert ChromiumWebbrowser.fallback_paths() == [
            "/Applications/Chromium.app/Contents/MacOS/Chromium",
            "/Users/user/Applications/Chromium.app/Contents/MacOS/Chromium",
            "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome",
            "/Users/user/Applications/Google Chrome.app/Contents/MacOS/Google Chrome",
        ]

    def test_other(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("streamlink.webbrowser.chromium.is_win32", False)
        monkeypatch.setattr("streamlink.webbrowser.chromium.is_darwin", False)
        assert ChromiumWebbrowser.fallback_paths() == []


def test_default_args():
    webbrowser = ChromiumWebbrowser()
    assert "--password-store=basic" in webbrowser.arguments
    assert "--use-mock-keychain" in webbrowser.arguments
    assert "--headless=new" not in webbrowser.arguments
    assert not any(arg.startswith("--remote-debugging-host") for arg in webbrowser.arguments)
    assert not any(arg.startswith("--remote-debugging-port") for arg in webbrowser.arguments)
    assert not any(arg.startswith("--user-data-dir") for arg in webbrowser.arguments)


@pytest.mark.trio()
@pytest.mark.parametrize("host", [pytest.param("127.0.0.1", id="ipv4"), pytest.param("::1", id="ipv6")])
@pytest.mark.parametrize("port", [pytest.param(None, id="default-port"), pytest.param(1234, id="custom-port")])
@pytest.mark.parametrize("headless", [pytest.param(True, id="headless"), pytest.param(False, id="not-headless")])
async def test_launch(
    monkeypatch: pytest.MonkeyPatch,
    mock_clock,
    webbrowser_launch,
    host: str,
    port: int | None,
    headless: bool,
):
    async def fake_find_free_port(_):
        await trio.lowlevel.checkpoint()
        return 1234

    monkeypatch.setattr("streamlink.webbrowser.chromium.find_free_port_ipv4", fake_find_free_port)
    monkeypatch.setattr("streamlink.webbrowser.chromium.find_free_port_ipv6", fake_find_free_port)

    webbrowser = ChromiumWebbrowser(host=host, port=port)

    process: trio.Process
    async with webbrowser_launch(webbrowser=webbrowser, headless=headless, timeout=999) as (_nursery, process):
        assert process.poll() is None, "process is still running"
        assert f"--remote-debugging-host={host}" in process.args
        assert "--remote-debugging-port=1234" in process.args
        assert ("--headless=new" in process.args) is headless
        param_user_data_dir = next(  # pragma: no branch
            (arg for arg in process.args if arg.startswith("--user-data-dir=")),
            None,
        )
        assert param_user_data_dir is not None

        user_data_dir = Path(param_user_data_dir[len("--user-data-dir=") :])
        assert user_data_dir.exists()

        # turn the 0.5s sleep() call at the end into a 0.5ms sleep() call
        # autojump_clock=0 would trigger the process's kill() fallback immediately and raise a warning
        mock_clock.rate = 1000

    assert process.poll() == (1 if is_win32 else -SIGTERM), "Process has been terminated"
    assert not user_data_dir.exists()


@pytest.mark.parametrize(
    ("host", "port", "address"),
    [
        pytest.param("127.0.0.1", 1234, "http://127.0.0.1:1234/json/version", id="IPv4"),
        pytest.param("::1", 1234, "http://[::1]:1234/json/version", id="IPv6"),
    ],
)
@pytest.mark.parametrize(
    "session",
    [
        pytest.param({"http-proxy": "http://localhost:4321/"}, id="with-proxy"),
        pytest.param({}, id="without-proxy"),
    ],
    indirect=True,
)
@pytest.mark.parametrize(
    ("num", "raises"),
    [
        pytest.param(10, nullcontext(), id="Success"),
        pytest.param(11, pytest.raises(PluginError), id="Timeout/Failure"),
    ],
)
def test_get_websocket_address(
    monkeypatch: pytest.MonkeyPatch,
    requests_mock: rm.Mocker,
    session: Streamlink,
    host: str,
    port: int,
    address: str,
    num: int,
    raises: nullcontext,
):
    monkeypatch.setattr("time.sleep", lambda _: None)
    hostaddr = f"[{host}]" if ":" in host else host

    payload = {
        "Browser": "Chrome/114.0.5735.133",
        "Protocol-Version": "1.3",
        "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
        "V8-Version": "11.4.183.23",
        "WebKit-Version": "537.36 (@fbfa2ce68d01b2201d8c667c2e73f648a61c4f4a)",
        "webSocketDebuggerUrl": f"ws://{hostaddr}:{port}/devtools/browser/some-uuid4",
    }

    responses: list[dict[str, Any]] = [{"exc": Timeout()} for _ in range(num)]
    responses.append({"json": payload})
    mock = requests_mock.register_uri("GET", address, responses)

    webbrowser = ChromiumWebbrowser(host=host, port=port)
    with raises:
        assert webbrowser.get_websocket_url(session) == f"ws://{hostaddr}:{port}/devtools/browser/some-uuid4"
        assert mock.called
        assert mock.last_request
        assert not mock.last_request.proxies.get("http")



from __future__ import annotations

import re
from collections.abc import Generator
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Generic, TypeVar


try:
    # noinspection PyProtectedMember
    from versioningit.onbuild import SetuptoolsFileProvider  # type: ignore[attr-defined]
except ImportError:  # pragma: no cover

    @dataclass
    class SetuptoolsFileProvider:  # type: ignore[no-redef]
        build_dir: Path


# noinspection PyUnusedLocal
def onbuild(
    *,
    is_source: bool,
    template_fields: dict[str, Any],
    params: dict[str, Any],
    # backward compatibility for `versioningit <3.0.0`
    build_dir: str | Path | None = None,
    file_provider: SetuptoolsFileProvider | None = None,
):
    """
    Remove the ``versioningit`` build-requirement from Streamlink's source distribution.
    Also set the static version string in the :mod:`streamlink._version` module when building the sdist/bdist.

    The version string already gets set by ``versioningit`` when building, so the sdist doesn't need to have
    ``versioningit`` added as a build-requirement. Previously, the generated version string was only applied
    to the :mod:`streamlink._version` module while ``versioningit`` was still set as a build-requirement.

    This custom onbuild hook gets called via the ``tool.versioningit.onbuild`` config in ``pyproject.toml``,
    since ``versioningit`` does only support modifying one file via its default onbuild hook configuration.
    """

    base_dir: Path
    if file_provider:
        base_dir = file_provider.build_dir.resolve()
    elif build_dir:  # pragma: no cover
        base_dir = Path(build_dir).resolve()
    else:  # pragma: no cover
        raise RuntimeError("Missing file_provider or build_dir")

    pkg_dir: Path = base_dir / "src" if is_source else base_dir
    version: str = template_fields["version"]
    cmproxy: Proxy[str]

    # Remove versioningit from ``build-system.requires`` in ``pyproject.toml``
    if is_source:
        with update_file(base_dir / "pyproject.toml") as cmproxy:
            cmproxy.set(
                re.sub(
                    r"^(\s*)(\"versioningit\b.+?\",).*$",
                    "\\1# \\2",
                    cmproxy.get(),
                    flags=re.MULTILINE,
                    count=1,
                ),
            )

    # Set the static version string that gets passed directly to setuptools via ``setup.py``.
    # This is much easier compared to adding the ``project.version`` field and removing "version" from ``project.dynamic``
    # in ``pyproject.toml``.
    if is_source:
        with update_file(base_dir / "setup.py") as cmproxy:
            cmproxy.set(
                re.sub(
                    r"^(\s*)# (version=\"\",).*$",
                    f'\\1version="{version}",',
                    cmproxy.get(),
                    flags=re.MULTILINE,
                    count=1,
                ),
            )

    # Overwrite the entire ``streamlink._version`` module
    with update_file(pkg_dir / "streamlink" / "_version.py") as cmproxy:
        cmproxy.set(f'__version__ = "{version}"\n')


TProxyItem = TypeVar("TProxyItem")


class Proxy(Generic[TProxyItem]):
    def __init__(self, data: TProxyItem):
        self._data = data

    def get(self) -> TProxyItem:
        return self._data

    def set(self, data: TProxyItem) -> None:
        self._data = data


@contextmanager
def update_file(file: Path) -> Generator[Proxy[str], None, None]:
    with file.open("r+", encoding="utf-8") as fh:
        proxy = Proxy(fh.read())
        yield proxy
        fh.seek(0)
        fh.write(proxy.get())
        fh.truncate()



from streamlink.plugins.btv import BTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlBTV(PluginCanHandleUrl):
    __plugin__ = BTV

    should_match = [
        "https://btvplus.bg/live",
        "https://btvplus.bg/live/",
        "https://www.btvplus.bg/live/",
    ]



from io import BytesIO
from unittest.mock import Mock, call

import pytest

from streamlink import Streamlink
from streamlink.stream.file import FileStream


class TestFileStream:
    @pytest.fixture(autouse=True)
    def mock_open(self, monkeypatch: pytest.MonkeyPatch) -> Mock:
        mock_open = Mock(return_value=BytesIO(b"foo"))
        monkeypatch.setattr("builtins.open", mock_open)

        return mock_open

    def test_open_path(self, session: Streamlink, mock_open: Mock):
        stream = FileStream(session, path="/test/path")
        streamio = stream.open()
        assert mock_open.call_args_list == [call("/test/path", "rb")]
        assert streamio.read() == b"foo"

        streamio.close()

    def test_open_fileobj(self, session: Streamlink, mock_open: Mock):
        fileobj = BytesIO(b"bar")
        stream = FileStream(session, fileobj=fileobj)
        streamio = stream.open()
        assert streamio is fileobj
        assert mock_open.call_args_list == []
        assert streamio.read() == b"bar"

        streamio.close()



import re
import shutil
from pathlib import Path

import pytest

from build_backend.onbuild import onbuild


try:
    # noinspection PyProtectedMember
    from versioningit.onbuild import SetuptoolsFileProvider  # type: ignore[attr-defined]
except ImportError:  # pragma: no cover
    _HAS_ONBUILD_FILE_PROVIDER = False
else:
    _HAS_ONBUILD_FILE_PROVIDER = True


PROJECT_ROOT = Path(__file__).parents[1]


@pytest.fixture()
def template_fields(request: pytest.FixtureRequest) -> dict:
    template_fields = {
        "version": "1.2.3+fake",
    }
    template_fields.update(getattr(request, "param", {}))

    return template_fields


@pytest.fixture(autouse=True)
def build(request: pytest.FixtureRequest, tmp_path: Path, template_fields: dict) -> Path:
    param = getattr(request, "param", {})
    is_source = param.get("is_source", True)
    pkg_dir = tmp_path / "src" if is_source else tmp_path

    (pkg_dir / "streamlink").mkdir(parents=True)
    shutil.copy(PROJECT_ROOT / "pyproject.toml", tmp_path / "pyproject.toml")
    shutil.copy(PROJECT_ROOT / "setup.py", tmp_path / "setup.py")

    # Lookup for the path from the root source directory
    if Path(PROJECT_ROOT / "src" / "streamlink" / "_version.py").exists():
        shutil.copy(PROJECT_ROOT / "src" / "streamlink" / "_version.py", pkg_dir / "streamlink" / "_version.py")
    else:  # pragma: no cover
        # We didn't find it, use the build location
        shutil.copy(PROJECT_ROOT / "streamlink" / "_version.py", pkg_dir / "streamlink" / "_version.py")

    options = dict(
        is_source=is_source,
        template_fields=template_fields,
        params={},
    )
    if _HAS_ONBUILD_FILE_PROVIDER:
        options["file_provider"] = SetuptoolsFileProvider(build_dir=tmp_path)
    else:  # pragma: no cover
        options["build_dir"] = tmp_path

    onbuild(**options)

    return tmp_path


@pytest.mark.parametrize("build", [pytest.param({"is_source": True}, id="is_source=True")], indirect=True)
def test_sdist(build: Path):
    assert re.search(
        r"^(\s*)# (\"versioningit\b.+?\",).*$",
        (build / "pyproject.toml").read_text(encoding="utf-8"),
        re.MULTILINE,
    ), "versioningit is not a build-requirement"
    assert re.search(
        # Check for any version string, and not just the fake one,
        # because tests can be run from the sdist where the version string was already set.
        # The onbuild hook does only replace the template in `setup.py`.
        r"^(\s*)(version=\"[^\"]+\",).*$",
        (build / "setup.py").read_text(encoding="utf-8"),
        re.MULTILINE,
    ), "setup() call defines a static version string"
    assert (
        (build / "src" / "streamlink" / "_version.py").read_text(encoding="utf-8")
        == '__version__ = "1.2.3+fake"\n'
    ), "streamlink._version exports a static version string"  # fmt: skip


@pytest.mark.parametrize("build", [pytest.param({"is_source": False}, id="is_source=False")], indirect=True)
def test_bdist(build: Path):
    assert (
        (build / "pyproject.toml").read_text(encoding="utf-8") == (PROJECT_ROOT / "pyproject.toml").read_text(encoding="utf-8")
    ), "Doesn't touch pyproject.toml (irrelevant for non-sdist)"  # fmt: skip
    assert (
        (build / "setup.py").read_text(encoding="utf-8") == (PROJECT_ROOT / "setup.py").read_text(encoding="utf-8")
    ), "Doesn't touch setup.py (irrelevant for non-sdist)"  # fmt: skip
    assert (
        (build / "streamlink" / "_version.py").read_text(encoding="utf-8") == '__version__ = "1.2.3+fake"\n'
    ), "streamlink._version exports a static version string"  # fmt: skip



from streamlink.plugins.cnbce import CNBCE
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlCNBCE(PluginCanHandleUrl):
    __plugin__ = CNBCE

    should_match = [
        "https://www.cnbce.com/canli-yayin",
    ]



from streamlink.plugins.steam import SteamBroadcastPlugin
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlSteamBroadcastPlugin(PluginCanHandleUrl):
    __plugin__ = SteamBroadcastPlugin

    should_match = [
        "https://steamcommunity.com/broadcast/watch/12432432",
        "http://steamcommunity.com/broadcast/watch/342342",
        "https://steam.tv/dota2",
        "http://steam.tv/dota2",
    ]

    should_not_match = [
        "http://steamcommunity.com/broadcast",
        "https://steamcommunity.com",
    ]



from streamlink.plugins.okru import OKru
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlOKru(PluginCanHandleUrl):
    __plugin__ = OKru

    should_match = [
        ("default", "https://ok.ru/live/ID"),
        ("default", "https://ok.ru/video/ID"),
        ("default", "https://ok.ru/videoembed/ID"),
        ("default", "https://www.ok.ru/video/ID"),
        ("mobile", "https://m.ok.ru/video/ID"),
        ("mobile", "https://mobile.ok.ru/video/ID"),
    ]



from __future__ import annotations

import sys

# noinspection PyProtectedMember
from ctypes import Structure, _SimpleCData  # noqa: PLC2701
from ctypes.wintypes import DWORD, WCHAR, WORD
from io import TextIOWrapper
from types import ModuleType
from typing import ClassVar, Generic, TypeVar
from unittest.mock import ANY, Mock, call

import pytest

from streamlink_cli.console.windows import COORD, WindowsConsole


_TCTypesType = TypeVar("_TCTypesType")


class _CTypesComparable(Generic[_TCTypesType]):
    """Allow comparing ctypes data types with built-in types for equality"""

    _type: ClassVar[type]

    def __init__(self, data: _TCTypesType):
        self.data: tuple = self._get_data(data)

    def _get_data(self, data: _TCTypesType) -> tuple:  # pragma: no cover
        raise NotImplementedError

    def __eq__(self, other):
        if isinstance(other, self._type):
            return self.__eq__(type(self)(other))
        if not isinstance(other, type(self)):  # pragma: no cover
            return False

        return self.data == other.data

    def __hash__(self):  # pragma: no cover
        return super().__hash__()


class EqSimpleCData(_CTypesComparable[_SimpleCData]):
    _type: ClassVar = _SimpleCData

    def _get_data(self, data: _SimpleCData) -> tuple:
        # noinspection PyProtectedMember
        return data._type_, data.value  # type: ignore[attr-defined]


class EqStructure(_CTypesComparable[Structure]):
    _type: ClassVar = Structure

    def _get_data(self, data: Structure) -> tuple:
        # noinspection PyProtectedMember
        return tuple(getattr(data, item) for (item, *_) in data._fields_)


@pytest.fixture(autouse=True)
def _mock_byref(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("streamlink_cli.console.windows.byref", Mock(side_effect=lambda obj, *_, **__: obj))


@pytest.fixture(autouse=True)
def mock_windll(monkeypatch: pytest.MonkeyPatch):
    mock_windll = Mock()
    monkeypatch.setattr("ctypes.windll", mock_windll, raising=False)

    return mock_windll


def test_no_windll(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setitem(sys.modules, "ctypes", ModuleType("ctypes"))
    assert WindowsConsole() is None


@pytest.mark.parametrize(
    ("method", "function"),
    [
        ("get_std_handle", "GetStdHandle"),
        ("get_console_mode", "GetConsoleMode"),
        ("get_console_screen_buffer_info", "GetConsoleScreenBufferInfo"),
        ("set_console_cursor_position", "SetConsoleCursorPosition"),
        ("fill_console_output_attribute", "FillConsoleOutputAttribute"),
        ("fill_console_output_character_w", "FillConsoleOutputCharacterW"),
    ],
)
def test_functions(mock_windll: Mock, method: str, function: str):
    windows_console = WindowsConsole()
    assert isinstance(windows_console, WindowsConsole)
    assert getattr(windows_console, method).method is getattr(mock_windll.kernel32, function)


def test_call_success_error(monkeypatch: pytest.MonkeyPatch, mock_windll: Mock):
    windows_console = WindowsConsole()
    monkeypatch.setattr(windows_console.set_console_cursor_position, "method", Mock(return_value=False))
    mock_windll.kernel32.GetLastError.return_value = 87

    with pytest.raises(OSError) as exc_info:  # noqa: PT011
        windows_console.set_console_cursor_position(123, 456)
    assert str(exc_info.value) == "Error while calling kernel32.SetConsoleCursorPosition (last_error=0x57)"


@pytest.mark.parametrize(
    ("stream", "expected"),
    [
        pytest.param(None, -11, id="None"),
        pytest.param(sys.stdout, -11, id="stdout"),
        pytest.param(sys.stderr, -12, id="stderr"),
    ],
)
def test_std_handle(mock_windll: Mock, stream: TextIOWrapper | None, expected: int):
    windows_console = WindowsConsole(stream)
    assert mock_windll.kernel32.GetStdHandle.call_args_list == [call(expected)]
    assert windows_console.handle is mock_windll.kernel32.GetStdHandle.return_value


@pytest.mark.parametrize(
    ("value", "success", "expected"),
    [
        pytest.param(0, False, False, id="error"),
        pytest.param(3, True, False, id="no-virtual-terminal-processing"),
        pytest.param(7, True, True, id="virtual-terminal-processing"),
    ],
)
def test_supports_virtual_terminal_processing(mock_windll: Mock, value: int, success: bool, expected: bool):
    def fake_get_console_mode(_handle, mode):
        mode.value = value

        return success

    mock_windll.kernel32.GetConsoleMode.side_effect = fake_get_console_mode
    mock_windll.kernel32.GetLastError.return_value = 87

    windows_console = WindowsConsole()
    assert windows_console.supports_virtual_terminal_processing() == expected
    assert mock_windll.kernel32.GetConsoleMode.call_args_list == [call(mock_windll.kernel32.GetStdHandle.return_value, ANY)]


def test_clear_line(mock_windll: Mock):
    def fake_get_console_screen_buffer_info(_handle, console_screen_buffer_info):
        console_screen_buffer_info.dwSize.X = 144
        console_screen_buffer_info.dwSize.Y = 42
        console_screen_buffer_info.dwCursorPosition.X = 20
        console_screen_buffer_info.dwCursorPosition.Y = 15
        console_screen_buffer_info.wAttributes = 0

        return True

    mock_windll.kernel32.GetConsoleScreenBufferInfo.side_effect = fake_get_console_screen_buffer_info

    windows_console = WindowsConsole()
    windows_console.clear_line()
    assert mock_windll.kernel32.GetConsoleScreenBufferInfo.call_args_list == [
        call(mock_windll.kernel32.GetStdHandle.return_value, ANY),
    ]
    assert mock_windll.kernel32.FillConsoleOutputCharacterW.call_args_list == [
        call(
            mock_windll.kernel32.GetStdHandle.return_value,
            EqSimpleCData(WCHAR(" ")),
            EqSimpleCData(DWORD(144)),
            EqStructure(COORD(0, 15)),
            EqSimpleCData(DWORD(0)),
        ),
    ]
    assert mock_windll.kernel32.FillConsoleOutputAttribute.call_args_list == [
        call(
            mock_windll.kernel32.GetStdHandle.return_value,
            EqSimpleCData(WORD(0)),
            EqSimpleCData(DWORD(144)),
            EqStructure(COORD(0, 15)),
            EqSimpleCData(DWORD(0)),
        ),
    ]
    assert mock_windll.kernel32.SetConsoleCursorPosition.call_args_list == [
        call(
            mock_windll.kernel32.GetStdHandle.return_value,
            EqStructure(COORD(0, 15)),
        ),
    ]



from streamlink.plugins.pluto import Pluto
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlPluto(PluginCanHandleUrl):
    __plugin__ = Pluto

    should_match_groups = [
        (
            (
                "live",
                "https://pluto.tv/en/live-tv/61409f8d6feb30000766b675",
            ),
            {
                "id": "61409f8d6feb30000766b675",
            },
        ),
        (
            (
                "series",
                "https://pluto.tv/en/on-demand/series/5e00cd538e67b0dcb2cf3bcd/season/1/episode/60dee91cfc802600134b886d",
            ),
            {
                "id_s": "5e00cd538e67b0dcb2cf3bcd",
                "id_e": "60dee91cfc802600134b886d",
            },
        ),
        (
            (
                "movies",
                "https://pluto.tv/en/on-demand/movies/600545d1813b2d001b686fa9",
            ),
            {
                "id": "600545d1813b2d001b686fa9",
            },
        ),
    ]

    should_not_match = [
        "https://pluto.tv/live-tv",
        "https://pluto.tv/en/live-tv/61409f8d6feb30000766b675/details",
        "https://pluto.tv/en/on-demand/series/5e00cd538e67b0dcb2cf3bcd/details/season/1",
        "https://pluto.tv/en/on-demand/movies/600545d1813b2d001b686fa9/details",
    ]



from __future__ import annotations

from collections.abc import Awaitable, Callable
from contextlib import nullcontext
from typing import TYPE_CHECKING, cast
from unittest.mock import ANY, AsyncMock, Mock, call

import pytest
import trio
from trio.testing import wait_all_tasks_blocked

from streamlink.compat import ExceptionGroup
from streamlink.session import Streamlink
from streamlink.webbrowser.cdp.client import CDPClient, CDPClientSession, RequestPausedHandler
from streamlink.webbrowser.cdp.connection import CDPConnection, CDPSession
from streamlink.webbrowser.cdp.devtools.fetch import RequestPaused
from streamlink.webbrowser.cdp.devtools.target import SessionID, TargetID
from streamlink.webbrowser.cdp.exceptions import CDPError
from tests.webbrowser.cdp import FakeWebsocketConnection


if TYPE_CHECKING:
    from typing_extensions import TypeAlias


TAsyncHandler: TypeAlias = "AsyncMock | Callable[[CDPClientSession, RequestPaused], Awaitable]"


def async_handler(*args, **kwargs):
    return cast(TAsyncHandler, AsyncMock(*args, **kwargs))


@pytest.fixture()
def chromium_webbrowser(monkeypatch: pytest.MonkeyPatch):
    # noinspection PyUnusedLocal
    def mock_launch(*args, **kwargs):
        return trio.open_nursery()

    mock_chromium_webbrowser = Mock(
        launch=Mock(side_effect=mock_launch),
        get_websocket_url=Mock(return_value="ws://localhost:1234/fake"),
    )
    mock_chromium_webbrowser_class = Mock(return_value=mock_chromium_webbrowser)
    monkeypatch.setattr("streamlink.webbrowser.cdp.client.ChromiumWebbrowser", mock_chromium_webbrowser_class)
    return mock_chromium_webbrowser


@pytest.fixture()
async def cdp_client(
    request: pytest.FixtureRequest,
    session: Streamlink,
    chromium_webbrowser: Mock,
    websocket_connection: FakeWebsocketConnection,
):
    params = getattr(request, "param", {})
    headless = params.get("headless", False)
    async with CDPClient.run(session, headless=headless) as cdp_client:
        yield cdp_client


@pytest.fixture()
def cdp_client_session(request: pytest.FixtureRequest, cdp_client: CDPClient):
    target_id = TargetID("01234")
    session_id = SessionID("56789")
    session = cdp_client.cdp_connection.sessions[session_id] = CDPSession(
        cdp_client.cdp_connection.websocket,
        target_id=target_id,
        session_id=session_id,
        cmd_timeout=cdp_client.cdp_connection.cmd_timeout,
    )
    fail_unhandled_requests = getattr(request, "param", False)
    return CDPClientSession(cdp_client, session, fail_unhandled_requests)


class TestLaunch:
    @pytest.fixture()
    def cdp_client(self):
        return Mock()

    @pytest.fixture(autouse=True)
    def mock_run(self, monkeypatch: pytest.MonkeyPatch, cdp_client: Mock):
        mock_run = Mock(
            return_value=Mock(
                __aenter__=AsyncMock(return_value=cdp_client),
                __aexit__=AsyncMock(),
            ),
        )
        monkeypatch.setattr(CDPClient, "run", mock_run)
        return mock_run

    @pytest.fixture(autouse=True)
    def _mock_launch(self, request: pytest.FixtureRequest, session: Streamlink, mock_run, cdp_client: Mock):
        result = object()
        mock_runner = AsyncMock(return_value=result)
        with getattr(request, "param", nullcontext()):
            assert CDPClient.launch(session, mock_runner) is result
            assert mock_runner.await_args_list == [call(cdp_client)]

    @pytest.mark.parametrize(
        ("session", "options"),
        [
            pytest.param(
                {},
                dict(executable=None, timeout=20.0, cdp_host=None, cdp_port=None, cdp_timeout=2.0, headless=False),
                id="Default options",
            ),
            pytest.param(
                {
                    "webbrowser-executable": "foo",
                    "webbrowser-timeout": 123.45,
                    "webbrowser-cdp-host": "::1",
                    "webbrowser-cdp-port": 1234,
                    "webbrowser-cdp-timeout": 12.34,
                    "webbrowser-headless": True,
                },
                dict(executable="foo", timeout=123.45, cdp_host="::1", cdp_port=1234, cdp_timeout=12.34, headless=True),
                id="Custom options",
            ),
        ],
        indirect=["session"],
    )
    def test_options(self, session: Streamlink, mock_run: Mock, options: dict):
        assert mock_run.call_args_list == [call(session=session, **options)]

    # noinspection PyTestParametrized
    @pytest.mark.usefixtures("_mock_launch")
    @pytest.mark.parametrize(
        ("session", "_mock_launch"),
        [
            pytest.param(
                {"webbrowser": False},
                pytest.raises(CDPError, match=r"^The webbrowser API has been disabled by the user$"),
                id="Raises CDPError",
            ),
        ],
        indirect=["session", "_mock_launch"],
    )
    def test_disabled(self, session: Streamlink, mock_run):
        assert not mock_run.called


class TestRun:
    @pytest.mark.trio()
    async def test_no_session(
        self,
        session: Streamlink,
        chromium_webbrowser: Mock,
        cdp_client: CDPClient,
        websocket_connection: FakeWebsocketConnection,
    ):
        assert isinstance(cdp_client, CDPClient)
        assert isinstance(cdp_client.cdp_connection, CDPConnection)
        assert isinstance(cdp_client.nursery, trio.Nursery)
        assert chromium_webbrowser.launch.called
        assert chromium_webbrowser.get_websocket_url.call_args_list == [call(session)]
        assert websocket_connection.sent == []

    @pytest.mark.trio()
    @pytest.mark.parametrize("fail_unhandled_requests", [False, True])
    async def test_session(
        self,
        cdp_client: CDPClient,
        websocket_connection: FakeWebsocketConnection,
        fail_unhandled_requests,
    ):
        client_session = None

        async def new_session():
            nonlocal client_session
            async with cdp_client.session(fail_unhandled_requests=fail_unhandled_requests) as client_session:
                pass

        async with trio.open_nursery() as nursery:
            nursery.start_soon(new_session)
            await wait_all_tasks_blocked()
            nursery.start_soon(websocket_connection.sender.send, """{"id":0,"result":{"targetId":"01234"}}""")
            await wait_all_tasks_blocked()
            nursery.start_soon(websocket_connection.sender.send, """{"id":1,"result":{"sessionId":"56789"}}""")

        assert isinstance(client_session, CDPClientSession)
        assert isinstance(client_session.cdp_session, CDPSession)
        assert client_session._fail_unhandled == fail_unhandled_requests
        assert websocket_connection.sent == [
            """{"id":0,"method":"Target.createTarget","params":{"url":""}}""",
            """{"id":1,"method":"Target.attachToTarget","params":{"flatten":true,"targetId":"01234"}}""",
        ]


class TestEvaluate:
    @pytest.mark.trio()
    async def test_success(self, cdp_client_session: CDPClientSession, websocket_connection: FakeWebsocketConnection):
        result = None

        async def evaluate():
            nonlocal result
            result = await cdp_client_session.evaluate("new Promise(r=>r('foo'))")

        async with trio.open_nursery() as nursery:
            nursery.start_soon(evaluate)
            await wait_all_tasks_blocked()
            await websocket_connection.sender.send(
                """{"id":0,"sessionId":"56789","result":{"result":{"type":"string","value":"foo"}}}""",
            )

        assert result == "foo"

    @pytest.mark.trio()
    async def test_exception(self, cdp_client_session: CDPClientSession, websocket_connection: FakeWebsocketConnection):
        with pytest.raises(ExceptionGroup) as excinfo:  # noqa: PT012
            async with trio.open_nursery() as nursery:
                nursery.start_soon(cdp_client_session.evaluate, "/")

                await wait_all_tasks_blocked()
                # language=json
                await websocket_connection.sender.send("""
                    {"id":0, "sessionId":"56789", "result": {
                        "result": {"type": "object", "subclass": "error"},
                        "exceptionDetails": {
                            "exceptionId": 1,
                            "text": "Uncaught",
                            "lineNumber": 0,
                            "columnNumber": 0,
                            "exception": {
                                "type": "object",
                                "subtype": "error",
                                "className": "SyntaxError",
                                "description": "SyntaxError: Invalid regular expression: missing /"
                            }
                        }
                    }}
                """)

        assert excinfo.group_contains(CDPError, match="^SyntaxError: Invalid regular expression: missing /$")

    @pytest.mark.trio()
    async def test_error(self, cdp_client_session: CDPClientSession, websocket_connection: FakeWebsocketConnection):
        with pytest.raises(ExceptionGroup) as excinfo:  # noqa: PT012
            async with trio.open_nursery() as nursery:
                nursery.start_soon(cdp_client_session.evaluate, "new Error('foo')")

                await wait_all_tasks_blocked()
                # language=json
                await websocket_connection.sender.send("""
                    {"id":0, "sessionId":"56789", "result": {
                        "result": {
                            "type": "object",
                            "subtype": "error",
                            "className": "Error",
                            "description": "Error: foo\\n    at <anonymous>:1:1"
                        }
                    }}
                """)

        assert excinfo.group_contains(CDPError, match="^Error: foo\\n    at <anonymous>:1:1$")


class TestRequestPausedHandler:
    @pytest.mark.parametrize(
        ("url_pattern", "regex_pattern"),
        [
            pytest.param(
                r"abc?def?xyz",
                r"^abc.def.xyz$",
                id="Question mark",
            ),
            pytest.param(
                r"abc*def*xyz",
                r"^abc.+def.+xyz$",
                id="Star",
            ),
            pytest.param(
                r"^(.[a-z])\d$",
                r"^\^\(\.\[a\-z\]\)\\d\$$",
                id="Special characters",
            ),
            pytest.param(
                r"abc\?def\*xyz",
                r"^abc\?def\*xyz$",
                id="Escaped question mark and star",
            ),
            pytest.param(
                r"abc\\?def\\*xyz",
                r"^abc\\\\.def\\\\.+xyz$",
                id="2 escape characters",
            ),
            pytest.param(
                r"abc\\\?def\\\*xyz",
                r"^abc\\\\\?def\\\\\*xyz$",
                id="3 escape characters",
            ),
            pytest.param(
                r"abc\\\\?def\\\\*xyz",
                r"^abc\\\\\\\\.def\\\\\\\\.+xyz$",
                id="4 escape characters",
            ),
            pytest.param(
                r"abc\\\\\?def\\\\\*xyz",
                r"^abc\\\\\\\\\?def\\\\\\\\\*xyz$",
                id="5 escape characters",
            ),
            pytest.param(
                r"http://*.name.tld/foo\?bar=baz",
                r"^http://.+\.name\.tld/foo\?bar=baz$",
                id="Typical URL pattern",
            ),
        ],
    )
    def test_url_pattern_to_regex_pattern(self, url_pattern: str, regex_pattern: str):
        assert RequestPausedHandler._url_pattern_to_regex_pattern(url_pattern).pattern == regex_pattern

    @pytest.mark.trio()
    async def test_client_registration(self, cdp_client_session: CDPClientSession):
        assert len(cdp_client_session._request_handlers) == 0
        cdp_client_session.add_request_handler(async_handler())
        cdp_client_session.add_request_handler(async_handler(), on_request=True)
        cdp_client_session.add_request_handler(async_handler(), url_pattern="foo")
        cdp_client_session.add_request_handler(async_handler(), url_pattern="foo", on_request=True)
        assert len(cdp_client_session._request_handlers) == 4
        assert all(request_handler.async_handler for request_handler in cdp_client_session._request_handlers)
        assert all(request_handler.url_pattern == "*" for request_handler in cdp_client_session._request_handlers[:2])
        assert all(request_handler.url_pattern == "foo" for request_handler in cdp_client_session._request_handlers[2:])
        assert not cdp_client_session._request_handlers[0].on_request
        assert not cdp_client_session._request_handlers[2].on_request
        assert cdp_client_session._request_handlers[1].on_request
        assert cdp_client_session._request_handlers[3].on_request

    @pytest.mark.parametrize(
        ("args", "matches"),
        [
            pytest.param(
                dict(async_handler=async_handler(), on_request=False),
                False,
                id="On response - Any URL",
            ),
            pytest.param(
                dict(async_handler=async_handler(), on_request=True),
                True,
                id="On request - Any URL",
            ),
            pytest.param(
                dict(async_handler=async_handler(), url_pattern="http://localhost/", on_request=True),
                True,
                id="Matching URL",
            ),
            pytest.param(
                dict(async_handler=async_handler(), url_pattern="http://l?c?l*/", on_request=True),
                True,
                id="Matching wildcard URL",
            ),
            pytest.param(
                dict(async_handler=async_handler(), url_pattern="http://other/", on_request=True),
                False,
                id="Non-matching URL",
            ),
        ],
    )
    def test_matches_request(self, args: dict, matches: bool):
        request = RequestPaused.from_json({
            "requestId": "request-1",
            "frameId": "frame-1",
            "request": {
                "url": "http://localhost/",
                "method": "GET",
                "headers": {},
                "initialPriority": "VeryHigh",
                "referrerPolicy": "strict-origin-when-cross-origin",
            },
            "resourceType": "Document",
        })
        request_handler = RequestPausedHandler(**args)
        assert request_handler.matches(request) is matches

    @pytest.mark.parametrize(
        ("args", "matches"),
        [
            pytest.param(
                dict(async_handler=async_handler(), on_request=False),
                True,
                id="On response - Any URL",
            ),
            pytest.param(
                dict(async_handler=async_handler(), on_request=True),
                False,
                id="On request - Any URL",
            ),
            pytest.param(
                dict(async_handler=async_handler(), url_pattern="http://localhost/", on_request=False),
                True,
                id="Matching URL",
            ),
            pytest.param(
                dict(async_handler=async_handler(), url_pattern="http://l?c?l*/", on_request=False),
                True,
                id="Matching wildcard URL",
            ),
            pytest.param(
                dict(async_handler=async_handler(), url_pattern="http://other/", on_request=False),
                False,
                id="Non-matching URL",
            ),
        ],
    )
    def test_matches_response(self, args: dict, matches: bool):
        request = RequestPaused.from_json({
            "requestId": "request-1",
            "frameId": "frame-1",
            "request": {
                "url": "http://localhost/",
                "method": "GET",
                "headers": {},
                "initialPriority": "VeryHigh",
                "referrerPolicy": "strict-origin-when-cross-origin",
            },
            "resourceType": "Document",
            "responseStatusCode": 200,
        })
        request_handler = RequestPausedHandler(**args)
        assert request_handler.matches(request) is matches


class TestNavigate:
    @pytest.mark.trio()
    async def test_detach(self, cdp_client_session: CDPClientSession, websocket_connection: FakeWebsocketConnection):
        async def navigate():
            async with cdp_client_session.navigate("https://foo"):
                pass  # pragma: no cover

        with pytest.raises(ExceptionGroup) as excinfo:  # noqa: PT012
            async with trio.open_nursery() as nursery:
                nursery.start_soon(navigate)

                await wait_all_tasks_blocked()
                await websocket_connection.sender.send(
                    """{"method":"Target.detachedFromTarget","params":{"sessionId":"unknown"}}""",
                )
                await wait_all_tasks_blocked()
                await websocket_connection.sender.send(
                    """{"method":"Target.detachedFromTarget","params":{"sessionId":"56789"}}""",
                )

        assert excinfo.group_contains(CDPError, match="^Target has been detached$")

    @pytest.mark.trio()
    async def test_error(self, cdp_client_session: CDPClientSession, websocket_connection: FakeWebsocketConnection):
        async def navigate():
            async with cdp_client_session.navigate("https://foo"):
                pass  # pragma: no cover

        with pytest.raises(ExceptionGroup) as excinfo:  # noqa: PT012
            async with trio.open_nursery() as nursery:
                nursery.start_soon(navigate)

                await wait_all_tasks_blocked()
                assert websocket_connection.sent == [
                    """{"id":0,"method":"Page.enable","params":{},"sessionId":"56789"}""",
                ]

                await websocket_connection.sender.send(
                    """{"id":0,"result":{},"sessionId":"56789"}""",
                )
                await wait_all_tasks_blocked()
                assert websocket_connection.sent == [
                    """{"id":0,"method":"Page.enable","params":{},"sessionId":"56789"}""",
                    """{"id":1,"method":"Page.navigate","params":{"url":"https://foo"},"sessionId":"56789"}""",
                ]

                await websocket_connection.sender.send(
                    """{"id":1,"result":{"frameId":"frame-id-1","errorText":"failure"},"sessionId":"56789"}""",
                )
                await wait_all_tasks_blocked()
                assert websocket_connection.sent == [
                    """{"id":0,"method":"Page.enable","params":{},"sessionId":"56789"}""",
                    """{"id":1,"method":"Page.navigate","params":{"url":"https://foo"},"sessionId":"56789"}""",
                    """{"id":2,"method":"Page.disable","sessionId":"56789"}""",
                ]

                await websocket_connection.sender.send(
                    """{"id":2,"result":{},"sessionId":"56789"}""",
                )

        assert excinfo.group_contains(CDPError, match="^Navigation error: failure$")

    @pytest.mark.trio()
    @pytest.mark.parametrize("cdp_client", [pytest.param({"headless": True}, id="headless")], indirect=True)
    async def test_loaded(
        self,
        cdp_client_session: CDPClientSession,
        websocket_connection: FakeWebsocketConnection,
        nursery: trio.Nursery,
    ):
        loaded = False

        async def navigate():
            nonlocal loaded
            async with cdp_client_session.navigate("https://foo") as frame_id:
                assert frame_id == "frame-id-1"
                await cdp_client_session.loaded(frame_id)
                loaded = True

        nursery.start_soon(navigate)

        expected_commands_sent = [
            """{"id":0,"method":"Runtime.evaluate","params":"""
            + """{"awaitPromise":false,"expression":"navigator.userAgent"},"sessionId":"56789"}""",
            """{"id":1,"method":"Network.setUserAgentOverride","params":{"userAgent":"A Chrome UA"},"sessionId":"56789"}""",
            """{"id":2,"method":"Page.enable","params":{},"sessionId":"56789"}""",
            """{"id":3,"method":"Page.navigate","params":{"url":"https://foo"},"sessionId":"56789"}""",
            """{"id":4,"method":"Page.disable","sessionId":"56789"}""",
        ]

        await wait_all_tasks_blocked()
        assert websocket_connection.sent == expected_commands_sent[0:1]
        await websocket_connection.sender.send(
            """{"id":0,"result":{"result":{"type":"string","value":"A HeadlessChrome UA"}},"sessionId":"56789"}""",
        )
        await wait_all_tasks_blocked()
        assert websocket_connection.sent == expected_commands_sent[0:2]
        await websocket_connection.sender.send(
            """{"id":1,"result":{},"sessionId":"56789"}""",
        )
        await wait_all_tasks_blocked()
        assert websocket_connection.sent == expected_commands_sent[0:3]
        await websocket_connection.sender.send(
            """{"id":2,"result":{},"sessionId":"56789"}""",
        )
        await wait_all_tasks_blocked()
        assert websocket_connection.sent == expected_commands_sent[0:4]
        await websocket_connection.sender.send(
            """{"id":3,"result":{"frameId":"frame-id-1"},"sessionId":"56789"}""",
        )
        await wait_all_tasks_blocked()
        await websocket_connection.sender.send(
            """{"method":"Page.frameStoppedLoading","params":{"frameId":"frame-id-unknown"},"sessionId":"56789"}""",
        )
        await wait_all_tasks_blocked()
        await websocket_connection.sender.send(
            """{"method":"Page.frameStoppedLoading","params":{"frameId":"frame-id-1"},"sessionId":"56789"}""",
        )
        await wait_all_tasks_blocked()
        assert websocket_connection.sent == expected_commands_sent[0:5]
        await websocket_connection.sender.send(
            """{"id":4,"result":{},"sessionId":"56789"}""",
        )

        assert loaded

    @pytest.mark.trio()
    @pytest.mark.parametrize(
        ("on_request", "fetch_enable_params"),
        [
            pytest.param(
                (False,),
                (
                    """{"handleAuthRequests":true,"patterns":[{"requestStage":"Response","urlPattern":"*"},"""
                    + """{"requestStage":"Response","urlPattern":"http://foo"}]}"""
                ),
                id="Single request handler, on response",
            ),
            pytest.param(
                (True,),
                (
                    """{"handleAuthRequests":true,"patterns":[{"requestStage":"Request","urlPattern":"*"},"""
                    + """{"requestStage":"Request","urlPattern":"http://foo"}]}"""
                ),
                id="Single request handler, on request",
            ),
            pytest.param(
                (False, False),
                (
                    """{"handleAuthRequests":true,"patterns":[{"requestStage":"Response","urlPattern":"*"},"""
                    + """{"requestStage":"Response","urlPattern":"http://foo"}]}"""
                ),
                id="Multiple request handlers, on response",
            ),
            pytest.param(
                (True, True),
                (
                    """{"handleAuthRequests":true,"patterns":[{"requestStage":"Request","urlPattern":"*"},"""
                    + """{"requestStage":"Request","urlPattern":"http://foo"}]}"""
                ),
                id="Multiple request handlers, on request",
            ),
            pytest.param(
                (False, True),
                (
                    """{"handleAuthRequests":true,"patterns":[{"requestStage":"Response","urlPattern":"*"},"""
                    + """{"requestStage":"Request","urlPattern":"*"},{"requestStage":"Response","urlPattern":"http://foo"},"""
                    + """{"requestStage":"Request","urlPattern":"http://foo"}]}"""
                ),
                id="Multiple request handlers, on response and on request",
            ),
            pytest.param(
                (True, False),
                (
                    """{"handleAuthRequests":true,"patterns":[{"requestStage":"Response","urlPattern":"*"},"""
                    + """{"requestStage":"Request","urlPattern":"*"},{"requestStage":"Response","urlPattern":"http://foo"},"""
                    + """{"requestStage":"Request","urlPattern":"http://foo"}]}"""
                ),
                id="Multiple request handlers, on request and on response",
            ),
        ],
    )
    async def test_fetch_enable(
        self,
        monkeypatch: pytest.MonkeyPatch,
        cdp_client_session: CDPClientSession,
        websocket_connection: FakeWebsocketConnection,
        nursery: trio.Nursery,
        on_request: tuple,
        fetch_enable_params: str,
    ):
        mock_on_fetch_request_paused = AsyncMock()
        monkeypatch.setattr(cdp_client_session, "_on_fetch_request_paused", mock_on_fetch_request_paused)

        for _on_request in on_request:
            cdp_client_session.add_request_handler(async_handler(), on_request=_on_request)
            cdp_client_session.add_request_handler(async_handler(), on_request=_on_request)
            cdp_client_session.add_request_handler(async_handler(), url_pattern="http://foo", on_request=_on_request)

        async def navigate():
            async with cdp_client_session.navigate("https://foo"):
                pass  # pragma: no cover

        assert not mock_on_fetch_request_paused.called

        nursery.start_soon(navigate)

        await wait_all_tasks_blocked()
        assert mock_on_fetch_request_paused.called
        assert websocket_connection.sent == [
            """{"id":0,"method":"Fetch.enable","params":""" + fetch_enable_params + ""","sessionId":"56789"}""",
        ]
        await websocket_connection.sender.send(
            """{"id":0,"result":{},"sessionId":"56789"}""",
        )

        await wait_all_tasks_blocked()
        assert websocket_connection.sent == [
            """{"id":0,"method":"Fetch.enable","params":""" + fetch_enable_params + ""","sessionId":"56789"}""",
            """{"id":1,"method":"Page.enable","params":{},"sessionId":"56789"}""",
        ]
        await websocket_connection.sender.send(
            """{"id":1,"result":{},"sessionId":"56789"}""",
        )

        await wait_all_tasks_blocked()
        assert websocket_connection.sent == [
            """{"id":0,"method":"Fetch.enable","params":""" + fetch_enable_params + ""","sessionId":"56789"}""",
            """{"id":1,"method":"Page.enable","params":{},"sessionId":"56789"}""",
            """{"id":2,"method":"Page.navigate","params":{"url":"https://foo"},"sessionId":"56789"}""",
        ]
        await websocket_connection.sender.send(
            """{"id":2,"result":{"frameId":"frame-id-1"},"sessionId":"56789"}""",
        )

        await wait_all_tasks_blocked()
        assert websocket_connection.sent == [
            """{"id":0,"method":"Fetch.enable","params":""" + fetch_enable_params + ""","sessionId":"56789"}""",
            """{"id":1,"method":"Page.enable","params":{},"sessionId":"56789"}""",
            """{"id":2,"method":"Page.navigate","params":{"url":"https://foo"},"sessionId":"56789"}""",
            """{"id":3,"method":"Page.disable","sessionId":"56789"}""",
        ]
        await websocket_connection.sender.send(
            """{"id":3,"result":{},"sessionId":"56789"}""",
        )

        await wait_all_tasks_blocked()
        assert websocket_connection.sent == [
            """{"id":0,"method":"Fetch.enable","params":""" + fetch_enable_params + ""","sessionId":"56789"}""",
            """{"id":1,"method":"Page.enable","params":{},"sessionId":"56789"}""",
            """{"id":2,"method":"Page.navigate","params":{"url":"https://foo"},"sessionId":"56789"}""",
            """{"id":3,"method":"Page.disable","sessionId":"56789"}""",
            """{"id":4,"method":"Fetch.disable","sessionId":"56789"}""",
        ]
        await websocket_connection.sender.send(
            """{"id":4,"result":{},"sessionId":"56789"}""",
        )


class TestRequestMethods:
    @pytest.fixture()
    def req_paused(self):
        return RequestPaused.from_json({
            "requestId": "request-1",
            "frameId": "frame-1",
            "request": {
                "url": "http://foo/",
                "method": "GET",
                "headers": {},
                "initialPriority": "VeryHigh",
                "referrerPolicy": "strict-origin-when-cross-origin",
            },
            "resourceType": "Document",
        })

    @pytest.mark.trio()
    async def test_continue_request(
        self,
        cdp_client_session: CDPClientSession,
        websocket_connection: FakeWebsocketConnection,
        req_paused: RequestPaused,
        nursery: trio.Nursery,
    ):
        nursery.start_soon(cdp_client_session.continue_request, req_paused, "http://bar", "POST", "data", {"a": "b", "c": "d"})

        await wait_all_tasks_blocked()
        assert "request-1" not in cdp_client_session._requests_handled
        assert websocket_connection.sent == [
            (
                """{"id":0,"method":"Fetch.continueRequest","params":"""
                + """{"headers":[{"name":"a","value":"b"},{"name":"c","value":"d"}],"method":"POST","""
                + """"postData":"ZGF0YQ==","requestId":"request-1","url":"http://bar"},"sessionId":"56789"}"""
            ),
        ]

        await websocket_connection.sender.send("""{"id":0,"result":{},"sessionId":"56789"}""")
        await wait_all_tasks_blocked()
        assert "request-1" in cdp_client_session._requests_handled

    @pytest.mark.trio()
    async def test_fail_request(
        self,
        cdp_client_session: CDPClientSession,
        websocket_connection: FakeWebsocketConnection,
        req_paused: RequestPaused,
        nursery: trio.Nursery,
    ):
        nursery.start_soon(cdp_client_session.fail_request, req_paused, "TimedOut")

        await wait_all_tasks_blocked()
        assert "request-1" not in cdp_client_session._requests_handled
        assert websocket_connection.sent == [
            (
                """{"id":0,"method":"Fetch.failRequest","params":"""
                + """{"errorReason":"TimedOut","requestId":"request-1"},"sessionId":"56789"}"""
            ),
        ]

        await websocket_connection.sender.send("""{"id":0,"result":{},"sessionId":"56789"}""")
        await wait_all_tasks_blocked()
        assert "request-1" in cdp_client_session._requests_handled

    @pytest.mark.trio()
    async def test_fulfill_request(
        self,
        cdp_client_session: CDPClientSession,
        websocket_connection: FakeWebsocketConnection,
        req_paused: RequestPaused,
        nursery: trio.Nursery,
    ):
        nursery.start_soon(cdp_client_session.fulfill_request, req_paused, 404, {"a": "b", "c": "d"}, "data")

        await wait_all_tasks_blocked()
        assert "request-1" not in cdp_client_session._requests_handled
        assert websocket_connection.sent == [
            (
                """{"id":0,"method":"Fetch.fulfillRequest","params":"""
                + """{"body":"ZGF0YQ==","requestId":"request-1","responseCode":404,"""
                + """"responseHeaders":[{"name":"a","value":"b"},{"name":"c","value":"d"}]},"sessionId":"56789"}"""
            ),
        ]

        await websocket_connection.sender.send("""{"id":0,"result":{},"sessionId":"56789"}""")
        await wait_all_tasks_blocked()
        assert "request-1" in cdp_client_session._requests_handled

    @pytest.mark.trio()
    async def test_alter_request(
        self,
        cdp_client_session: CDPClientSession,
        websocket_connection: FakeWebsocketConnection,
        req_paused: RequestPaused,
        nursery: trio.Nursery,
    ):
        async def alter_request():
            async with cdp_client_session.alter_request(req_paused, 404, {"a": "b", "c": "d"}) as cmproxy:
                assert cmproxy.body == ""
                cmproxy.body = "foo"

        nursery.start_soon(alter_request)

        await wait_all_tasks_blocked()
        assert "request-1" not in cdp_client_session._requests_handled
        assert websocket_connection.sent == [
            (
                """{"id":0,"method":"Fetch.fulfillRequest","params":"""
                + """{"body":"Zm9v","requestId":"request-1","responseCode":404,"""
                + """"responseHeaders":[{"name":"a","value":"b"},{"name":"c","value":"d"}]},"sessionId":"56789"}"""
            ),
        ]

        await websocket_connection.sender.send("""{"id":0,"result":{},"sessionId":"56789"}""")
        await wait_all_tasks_blocked()
        assert "request-1" in cdp_client_session._requests_handled

    @pytest.mark.trio()
    async def test_alter_request_response(
        self,
        cdp_client_session: CDPClientSession,
        websocket_connection: FakeWebsocketConnection,
        req_paused: RequestPaused,
        nursery: trio.Nursery,
    ):
        # turn the request into a response
        req_paused.response_status_code = 200

        async def alter_request():
            async with cdp_client_session.alter_request(req_paused, 404, {"a": "b", "c": "d"}) as cmproxy:
                assert cmproxy.body == "foo"
                assert cmproxy.response_code == 404
                assert cmproxy.response_headers == {"a": "b", "c": "d"}
                cmproxy.body = cmproxy.body.upper()
                cmproxy.response_code -= 3
                cmproxy.response_headers["c"] = "e"

        nursery.start_soon(alter_request)

        await wait_all_tasks_blocked()
        assert "request-1" not in cdp_client_session._requests_handled
        assert websocket_connection.sent == [
            """{"id":0,"method":"Fetch.getResponseBody","params":{"requestId":"request-1"},"sessionId":"56789"}""",
        ]

        await websocket_connection.sender.send(
            """{"id":0,"result":{"body":"Zm9v","base64Encoded":true},"sessionId":"56789"}""",
        )
        await wait_all_tasks_blocked()
        assert websocket_connection.sent == [
            """{"id":0,"method":"Fetch.getResponseBody","params":{"requestId":"request-1"},"sessionId":"56789"}""",
            (
                """{"id":1,"method":"Fetch.fulfillRequest","params":"""
                + """{"body":"Rk9P","requestId":"request-1","responseCode":401,"""
                + """"responseHeaders":[{"name":"a","value":"b"},{"name":"c","value":"e"}]},"sessionId":"56789"}"""
            ),
        ]

        await websocket_connection.sender.send("""{"id":1,"result":{},"sessionId":"56789"}""")
        await wait_all_tasks_blocked()
        assert "request-1" in cdp_client_session._requests_handled


class TestOnFetchRequestPaused:
    @pytest.mark.trio()
    async def test_unhandled_continue(
        self,
        monkeypatch: pytest.MonkeyPatch,
        cdp_client_session: CDPClientSession,
        websocket_connection: FakeWebsocketConnection,
        nursery: trio.Nursery,
    ):
        mock_fail_request = AsyncMock()
        mock_continue_request = AsyncMock()
        monkeypatch.setattr(cdp_client_session, "fail_request", mock_fail_request)
        monkeypatch.setattr(cdp_client_session, "continue_request", mock_continue_request)

        handler_foo = async_handler()
        handler_bar = async_handler()
        cdp_client_session.add_request_handler(handler_foo, url_pattern="http://foo/")
        cdp_client_session.add_request_handler(handler_bar, url_pattern="http://bar/")

        nursery.start_soon(cdp_client_session._on_fetch_request_paused)
        await wait_all_tasks_blocked()

        # language=json
        await websocket_connection.sender.send("""
            {
                "method": "Fetch.requestPaused",
                "params": {
                    "requestId": "request-1",
                    "frameId": "frame-1",
                    "request": {
                        "url": "http://bar/",
                        "method": "GET",
                        "headers": {},
                        "initialPriority": "VeryHigh",
                        "referrerPolicy": "strict-origin-when-cross-origin"
                    },
                    "resourceType": "Document",
                    "responseStatusCode": 200
                },
                "sessionId": "56789"
            }
        """)
        await wait_all_tasks_blocked()
        assert "request-1" not in cdp_client_session._requests_handled
        assert handler_foo.call_args_list == []
        assert handler_bar.call_args_list == [call(cdp_client_session, ANY)]
        assert isinstance(handler_bar.call_args_list[0][0][1], RequestPaused)
        assert mock_fail_request.call_args_list == []
        assert mock_continue_request.call_args_list == [call(ANY)]
        assert isinstance(mock_continue_request.call_args_list[0][0][0], RequestPaused)

    @pytest.mark.trio()
    async def test_unhandled_fail(
        self,
        monkeypatch: pytest.MonkeyPatch,
        cdp_client_session: CDPClientSession,
        websocket_connection: FakeWebsocketConnection,
        nursery: trio.Nursery,
    ):
        # make unhandled requests fail
        cdp_client_session._fail_unhandled = True

        mock_fail_request = AsyncMock()
        mock_continue_request = AsyncMock()
        monkeypatch.setattr(cdp_client_session, "fail_request", mock_fail_request)
        monkeypatch.setattr(cdp_client_session, "continue_request", mock_continue_request)

        handler_foo = async_handler()
        handler_bar = async_handler()
        cdp_client_session.add_request_handler(handler_foo, url_pattern="http://foo/")
        cdp_client_session.add_request_handler(handler_bar, url_pattern="http://bar/")

        nursery.start_soon(cdp_client_session._on_fetch_request_paused)
        await wait_all_tasks_blocked()

        # language=json
        await websocket_connection.sender.send("""
            {
                "method": "Fetch.requestPaused",
                "params": {
                    "requestId": "request-1",
                    "frameId": "frame-1",
                    "request": {
                        "url": "http://bar/",
                        "method": "GET",
                        "headers": {},
                        "initialPriority": "VeryHigh",
                        "referrerPolicy": "strict-origin-when-cross-origin"
                    },
                    "resourceType": "Document",
                    "responseStatusCode": 200
                },
                "sessionId": "56789"
            }
        """)
        await wait_all_tasks_blocked()
        assert "request-1" not in cdp_client_session._requests_handled
        assert handler_foo.call_args_list == []
        assert handler_bar.call_args_list == [call(cdp_client_session, ANY)]
        assert isinstance(handler_bar.call_args_list[0][0][1], RequestPaused)
        assert mock_fail_request.call_args_list == [call(ANY)]
        assert mock_continue_request.call_args_list == []
        assert isinstance(mock_fail_request.call_args_list[0][0][0], RequestPaused)

    @pytest.mark.trio()
    async def test_handled(
        self,
        monkeypatch: pytest.MonkeyPatch,
        cdp_client_session: CDPClientSession,
        websocket_connection: FakeWebsocketConnection,
        nursery: trio.Nursery,
    ):
        # make unhandled requests fail
        cdp_client_session._fail_unhandled = True

        mock_fail_request = AsyncMock()
        mock_continue_request = AsyncMock()
        monkeypatch.setattr(cdp_client_session, "fail_request", mock_fail_request)
        monkeypatch.setattr(cdp_client_session, "continue_request", mock_continue_request)

        def mock_handled(_cdp_client_session: CDPClientSession, request: RequestPaused):
            # pretend that we've called any of the request methods which register that the request was handled appropriately
            _cdp_client_session._requests_handled.add(request.request_id)

        handler_foo = async_handler()
        handler_bar = async_handler(side_effect=mock_handled)
        cdp_client_session.add_request_handler(handler_foo, url_pattern="http://foo/")
        cdp_client_session.add_request_handler(handler_bar, url_pattern="http://bar/")

        nursery.start_soon(cdp_client_session._on_fetch_request_paused)
        await wait_all_tasks_blocked()

        # language=json
        await websocket_connection.sender.send("""
            {
                "method": "Fetch.requestPaused",
                "params": {
                    "requestId": "request-1",
                    "frameId": "frame-1",
                    "request": {
                        "url": "http://bar/",
                        "method": "GET",
                        "headers": {},
                        "initialPriority": "VeryHigh",
                        "referrerPolicy": "strict-origin-when-cross-origin"
                    },
                    "resourceType": "Document",
                    "responseStatusCode": 200
                },
                "sessionId": "56789"
            }
        """)
        await wait_all_tasks_blocked()
        assert "request-1" in cdp_client_session._requests_handled
        assert handler_foo.call_args_list == []
        assert handler_bar.call_args_list == [call(cdp_client_session, ANY)]
        assert isinstance(handler_bar.call_args_list[0][0][1], RequestPaused)
        assert mock_fail_request.call_args_list == []
        assert mock_continue_request.call_args_list == []
